/*
 * bg_annotator.c
 * Copyright (C) 2019 Peter Belkner <pbelkner@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#if defined (_WIN32) // [
#include <locale.h>
#else // ] [
#include <wctype.h>
#include <wchar.h>
#include <fcntl.h>
#include <unistd.h>
#endif // ]
#include <bg.h>

#if defined (BG_ANNOTATOR) // [
///////////////////////////////////////////////////////////////////////////////
// NOTE:  the annotation phase is really needed because it may re-organize
//   the tree!
static bg_visitor_vtbl_t annotator_vtbl;

///////////////////////////////////////////////////////////////////////////////
int bg_annotator_create(bg_visitor_t *vis)
{
  /////////////////////////////////////////////////////////////////////////////
  vis->vtbl=&annotator_vtbl;
  vis->depth=0;

  /////////////////////////////////////////////////////////////////////////////
  return 0;
}

static void bg_annotator_destroy(bg_visitor_t *vis FFUNUSED)
{
}
#endif // ]

#if defined (BG_FILE) // [
///////////////////////////////////////////////////////////////////////////////
#if defined (BG_ANNOTATOR) // [
static int bg_annotator_dispatch_file(bg_visitor_t *vis FFUNUSED,
    bg_tree_t *tree)
#else // ] [
int bg_file_annotation_create(bg_tree_t *tree)
#endif // ]
{
  const bg_param_t *param=tree->param;
  bg_tree_t *parent=tree->parent;
  size_t len1,len2;
  ssize_t size;
  ffchar_t *tp;

#if defined (BG_OVERWRITE) // [
  if (param->overwrite)
    goto success;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  len1=parent&&parent->target.path?FFSTRLEN(parent->target.path):0u;
  len2=tree->source.basename?FFSTRLEN(tree->source.basename):0u;
  size=(len1?len1+1u:0u)+(len2?len2+1u:0u);
  tree->target.path=tp=malloc(size*sizeof tp[0]);

  if (!tree->target.path) {
    DMESSAGE("allocating output path");
    goto e_path;
  }

  if (parent&&parent->target.path) {
    FFSTRCPY(tp,parent->target.path);
    tp+=len1;

    if (tree->source.basename)
      *tp++=FFPATHSEP;
  }

  tree->target.basename=tp;

  if (tree->source.basename) {
    FFSTRCPY(tp,tree->source.basename);
    tp+=len2;
  }
  else
    *tp=0;

#if defined (BG_MUXER_PATH_TEMP) // [
  /////////////////////////////////////////////////////////////////////////////
#if defined (BG_TEMP_PREFIX) // [
  len2=FFSTRLEN(param->temp_prefix);
  tree->temp.path=tp=malloc((size+len2)*sizeof tp[0]);
#else // ] [
  tree->temp.path=tp=malloc((size+4)*sizeof tp[0]);
#endif // ]

  if (!tree->temp.path) {
    DMESSAGE("allocating temporary path");
    goto e_temp;
  }

  if (parent&&parent->target.path) {
    FFSTRCPY(tp,parent->target.path);
    tp+=len1;

    if (tree->source.basename)
      *tp++=FFPATHSEP;
  }

  tree->temp.basename=tp;
#if defined (BG_TEMP_PREFIX) // [
  FFSTRCPY(tp,param->temp_prefix);
  tp+=len2;
#else // ]
  *tp++=FFL('_');
  *tp++=FFL('_');
  *tp++=FFL('_');
  *tp++=FFL('_');
#endif // ]
  FFSTRCPY(tp,tree->target.basename);
#endif // ]

#if defined (BG_TEMP_PREFIX) // [
  if (ff_fexists(tree->temp.path)) {
    FFVMESSAGE(FFL("file \"%s\" exists. Either remove it or define another"
        " prefix for temporary files by means of option --temp-prefix"),
        tree->temp.path);
    goto e_temp_exists;
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
#if defined (BG_OVERWRITE) // [
success:
#endif // ]
  return 0;
//cleanup:
#if defined (BG_TEMP_PREFIX) // [
e_temp_exists:
#endif // ]
#if defined (BG_MUXER_PATH_TEMP) // [
  free(tree->temp.path);
e_temp:
#endif // ]
  free(tree->target.path);
e_path:
  return -1;
}

#if ! defined (BG_ANNOTATOR) // [
void bg_file_annotation_destroy(bg_tree_t *tree FFUNUSED)
{
#if defined (BG_MUXER_PATH_TEMP) // [
  free(tree->temp.path);
#endif // ]
  free(tree->target.path);
}
#endif // ]
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#if ! defined (FF_INPUT_OPEN_SFX) // [
#if defined (FF_OUTPUT_CALLBACK_SFX) // [
const ffchar_t *bg_track_out_sfx(const bg_track_t *track,
    const bg_tree_t *tree)
#else // ] [
static const ffchar_t *bg_track_out_sfx(const bg_track_t *track,
    const bg_tree_t *tree)
#endif // ]
{
#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
  bg_param_t *param=tree->param;
#endif // ]
  ffchar_t *sp,*sfx;

#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
  if (param->out.sfx&&*param->out.sfx) {
    // in case a sfx is provided choose it.
    return param->out.sfx;
  }
  else
#endif // ]
  if (0<=track->input.vi) {
    // in case of a video:
    if (track->input.audio.ctx) {
      // when transcoding we choose MKV as the container.
      return FFL("mkv");
    }
    else {
      // otherwise get the suffix from the input.
#if 0 // [
      sfx=FFSTRSTR(tree->source.path,FFL("."));

      return sfx&&*sfx?++sfx:FFL("mkv");
#else // ] [
      sfx=sp=FFSTRSTR(tree->source.path,FFL("."));

      // find the right-most dot '.'.
      while (sp) {
        sfx=++sp;
        sp=FFSTRSTR(sp,FFL("."));
      }

      // we always remux to mkv.
      if (!sfx)
        return FFL("mkv");
      else
        return FFL("mkv");
#endif // ]
    }
  }
  else if (track->input.audio.ctx) {
    // we always transcode to FLAC.
    return FFL("flac");
	}
  else {
    // get the suffix from the input.
    sfx=sp=FFSTRSTR(tree->source.path,FFL("."));

    // find the right-most dot '.'.
    while (sp) {
      sfx=++sp;
      sp=FFSTRSTR(sp,FFL("."));
    }

    if (!sfx)
      return FFL("mka");
    else if (!FFSTRCMP(FFL("mp3"),sfx))
      return sfx;
    else if (!FFSTRCMP(FFL("flac"),sfx))
      return sfx;
    else if (!FFSTRCMP(FFL("wv"),sfx))
      return sfx;
    else
      return FFL("mka");
  }
}
#endif // ]

static int bg_track_basename(bg_track_t *track, bg_tree_t *tree,
    ffchar_t **opath, size_t len)
{
  bg_param_t *param=tree->param;
  AVDictionary *metadata=track->input.fmt.ctx->metadata;
  const AVDictionaryEntry *title=av_dict_get(metadata,"TITLE",NULL,
      AV_DICT_IGNORE_SUFFIX);
#if defined (_WIN32) && ! defined (BG_WIN32_CREATE_LOCALE) // [
  const char *locale=NULL;
#endif // ]
  // NOTE: under Linux sizeof(wchar_t) is 4 bytes and under Windows
  //   it is 2 bytes.
  wchar_t *op;
  const ffchar_t *ipath,*sfx,*sp;

  if ((BG_FLAGS_EXT_RENAME&param->flags.extension)&&title) {
#if defined (_WIN32) // [
    if (opath) {
      // swprintf() in any case writes a trailing '\0', hence (1+len).
      swprintf(*opath,1+track->target.pfx.len,
          track->album.id<100?L"%02lu_":L"%lu_",track->album.id);
      wcscpy(*opath+track->target.pfx.len,track->target.title);
      free(track->target.title);
      track->target.title=NULL;
    }
    else {
      // claculate the length of the intermediate representation.
      len=MultiByteToWideChar(
        CP_UTF8,        // UINT                              CodePage,
        0ul,            // DWORD                             dwFlags,
        title->value,   // _In_NLS_string_(cbMultiByte)LPCCH lpMultiByteStr,
        strlen(title->value),
                        // int                               cbMultiByte,
        NULL,           // LPWSTR                            lpWideCharStr,
        0               // int                               cchWideChar
      );

      // allocate sufficient memory to hold the intermediate representatiion.
      track->target.title=malloc((len+1)*sizeof *track->target.title);

      if (!track->target.title) {
        DMESSAGE("allocating intermediate path");
        // set error.
        len=-1;
        goto e_path;
      }

      // transform into the intermediate representation.
      MultiByteToWideChar(
        CP_UTF8,        // UINT                              CodePage,
        0ul,            // DWORD                             dwFlags,
        title->value,   // _In_NLS_string_(cbMultiByte)LPCCH lpMultiByteStr,
        strlen(title->value),
                        // int                               cbMultiByte,
        track->target.title,
                        // LPWSTR                            lpWideCharStr,
        len             // int                               cchWideChar
      );

      // add a trailing '\0' to the intermediate representation.
      track->target.title[len]=L'\0';
      // do some transformations, i.e. replace some characters by '_'
      // and transform to lower case.
      op=track->target.title;
#if ! defined (BG_WIN32_CREATE_LOCALE) // [
      // we need to temporarily switch locale in order that towlower() works
      // as expected because unfortunately on our system _create_locale()
      // isn't availabe and hence we have to rely on towlower().
      locale=setlocale(LC_ALL,NULL);
      setlocale(LC_ALL,"");
#endif // ]

      while (*op) {
        switch (*op) {
        case L'.':
        case L'/':
        case L'\\':
        case L'(':
        case L')':
        case L'&':
        case L':':
        case L' ':
        case L'\'':
          if (track->target.title<op&&'_'==op[-1])
            memmove(op,op+1,(wcslen(op+1)+1)*sizeof *op);
          else {
            *op=L'_';
            op=CharNextW(op);
          }

          break;
        default:
#if defined (BG_WIN32_CREATE_LOCALE) // [
          *op=_towlower_l(*op,tree->param->locale);
#else // ] [
          // our system _create_locale() isn't availabe.
          *op=towlower(*op);
#endif // ]
          op=CharNextW(op);
          break;
        }
      }

      // if necessary shorten the intermediate representation.
      if (track->target.title<op&&'_'==op[-1])
        op[-1]=L'\0';

#if ! defined (BG_WIN32_CREATE_LOCALE) // [
      // we need to switch back locale.
      setlocale(LC_ALL,locale);
#endif // ]

#if defined (BG_ERROR_LENGTH) // [
      if (!*track->target.title) {
        free(track->target.title);
        track->target.title=NULL;
        goto e_length;
      }
#endif // ]

      len=swprintf(NULL,0,
          track->album.id<100?L"%02lu_":L"%lu_",track->album.id);
      track->target.pfx.len=len;
      len+=wcslen(track->target.title);
    }
#else // ] [
    if (opath) {
      // snprintf() in any case writes a trailing '\0', hence (1+len).
      // I.e. we assume enough memory is allocated for holding a
      // trailing '\0'.
      snprintf(*opath,1+track->target.pfx.len,
          track->album.id<100?"%02lu_":"%lu_",track->album.id);
      // we convert the intermediate representation back into utf-8.
      wcstombs(*opath+track->target.pfx.len,track->target.title,
          len-track->target.pfx.len);
      free(track->target.title);
      track->target.title=NULL;
    }
    else {
      // calculate the length for holding the intermediate representation.
      len=mbstowcs(NULL,title->value,0);
      // allocate sufficient memory  in order to hold the intermediate
      // representation.
      track->target.title=malloc((len+1)*sizeof track->target.title[0]);

      if (!track->target.title) {
        DMESSAGE("allocating intermediate path");
        // set error.
        len=-1;
        goto e_path;
      }

      // convert the title tag into the intermediate representation
      // including a trailing '\0'.
      mbstowcs(track->target.title,title->value,len+1);

      // do some transformations.
      op=track->target.title;

      while (*op) {
        switch (*op) {
        case L'.':
        case L'/':
        case L'\\':
        case L'(':
        case L')':
        case L'&':
        case L':':
        case L' ':
        case L'\'':
          if (track->target.title<op&&'_'==op[-1])
            memmove(op,op+1,(wcslen(op+1)+1)*sizeof *op);
          else {
            *op='_';
            ++op;
          }

          break;
        default:
          *op=towlower(*op);
          ++op;
          break;
        }
      }

      if (track->target.title<op&&'_'==op[-1])
        op[-1]=L'\0';

#if defined (BG_ERROR_LENGTH) // [
      if (!*track->target.title) {
        free(track->target.title);
        track->target.title=NULL;
        goto e_length;
      }
#endif // ]

      len=snprintf(NULL,0,track->album.id<100?"%02lu_":"%lu_",track->album.id);
      track->target.pfx.len=len;
      // add the length for holding the title (transformed to lower case.)
      len+=wcstombs(NULL,track->target.title,0);
    }
#endif // ]
  }
  else {
#if defined (BG_ERROR_LENGTH) // [
e_length:
#endif // ]
    if (tree->source.basename) {
      ipath=tree->source.basename;

      if (opath)
        memcpy(*opath,ipath,len*sizeof **opath);
      else {
        sfx=NULL;
        sp=FFSTRSTR(ipath,FFL("."));

        while (sp) {
          sfx=sp;
          sp=FFSTRSTR(++sp,FFL("."));
        }

        len=sfx?(size_t)(sfx-ipath):FFSTRLEN(ipath);
      }
    }
    else
      len=0;
  }
e_path:
  return len;
}

///////////////////////////////////////////////////////////////////////////////
#if defined (FF_INPUT_OPEN_SFX) // [
#if defined (BG_ANNOTATOR) // [
static int bg_annotator_dispatch_track(bg_visitor_t *vis FFUNUSED,
    bg_tree_t *tree, const ffchar_t *sfx)
#else // ] [
int bg_track_annotation_create(bg_tree_t *tree, const ffchar_t *sfx)
#endif // ]
#elif defined (BG_ANNOTATOR) // ] [
static int bg_annotator_dispatch_track(bg_visitor_t *vis FFUNUSED,
    bg_tree_t *tree)
#else // ] [
int bg_track_annotation_create(bg_tree_t *tree)
#endif // ]
{
  bg_param_t *param=tree->param;
  bg_track_t *track=&tree->track;
  bg_tree_t *parent=tree->parent;
  size_t len1,len2;
#if ! defined (FF_INPUT_OPEN_SFX) // [
  // before determining the suffix we need to have the file re-opened.
  const ffchar_t *sfx;
#endif // ]
  size_t len3;
  size_t size;
  ffchar_t *tp;

  /////////////////////////////////////////////////////////////////////////////
  if (param->overwrite) {
    // including the path separator!
    len1=tree->source.basename-tree->source.path;
#if 0 // [
    // pointer arithmetics takes into account the element's size!
    len1/=sizeof tree->source.path[0];
#endif // ]
    len2=bg_track_basename(track,tree,NULL,0);
    // before determining the suffix we need to have the file re-opened.
#if ! defined (FF_INPUT_OPEN_SFX) // [
    sfx=bg_track_out_sfx(track,tree);
#endif // ]
    len3=sfx?FFSTRLEN(sfx):0u;
#if 0 // [
    size=(len1?len1+1u:0u)+(len2?len2+1u:0u)+(len3?len3+2u:0u);
#else // ] [
    size=len1+(len2?len2+1u:0u)+(len3?len3+2u:0u);
#endif // ]
    tree->target.path=tp=malloc(size*sizeof tp[0]);

    if (!tree->target.path) {
      DMESSAGE("allocating output path");
      goto e_path;
    }

    // we copy the path separator!
    memcpy(tp,tree->source.path,len1*sizeof tree->source.path[0]);
    tp+=len1;

    tree->target.basename=tp;
    bg_track_basename(track,tree,&tp,len2);
    tp+=len2;
    *tp=0;

    if (sfx) {
      *tp++=FFL('.');
      FFSTRCPY(tp,sfx);
      tp+=len3;
    }

    *tp=0;
  }
  else {
    len1=parent&&parent->target.path?FFSTRLEN(parent->target.path):0u;
    len2=bg_track_basename(track,tree,NULL,0);
#if !defined (FF_INPUT_OPEN_SFX) // [
    // before determining the suffix we need to have the file re-opened.
    sfx=bg_track_out_sfx(track,tree);
#endif // ]
    len3=sfx?FFSTRLEN(sfx):0u;
    size=(len1?len1+1u:0u)+(len2?len2+1u:0u)+(len3?len3+2u:0u);
    tree->target.path=tp=malloc(size*sizeof tp[0]);

    if (!tree->target.path) {
      DMESSAGE("allocating output path");
      goto e_path;
    }

    if (parent&&parent->target.path) {
      FFSTRCPY(tp,parent->target.path);
      tp+=len1;
      *tp++=FFPATHSEP;
    }

    tree->target.basename=tp;
    bg_track_basename(track,tree,&tp,len2);
    tp+=len2;
    *tp=0;

    if (sfx) {
      *tp++=FFL('.');
      FFSTRCPY(tp,sfx);
      tp+=len3;
    }

    *tp=0;
  }

#if defined (BG_MUXER_PATH_TEMP) // [
  if (!param->overwrite&&!ff_fcmp(tree->source.path,tree->target.path)) {
    DMESSAGE("overwriting not permitted. Use option --overwrite");
    goto e_overwrite_target;
  }

  /////////////////////////////////////////////////////////////////////////////
#if defined (BG_TEMP_PREFIX) // [
  len2=FFSTRLEN(param->temp_prefix);
  tree->temp.path=tp=malloc((size+len2)*sizeof tp[0]);
#else // ] [
  tree->temp.path=tp=malloc((size+4)*sizeof tp[0]);
#endif // ]

  if (!tree->temp.path) {
    DMESSAGE("allocating temporary path");
    goto e_temp;
  }

#if 0 // [
  if (parent&&parent->target.path) {
    FFSTRCPY(tp,parent->target.path);
    tp+=len1;
    *tp++=FFPATHSEP;
  }
#else // ] [
  memcpy(tp,tree->target.path,len1*sizeof tp[0]);
  tp+=len1;

  if (FFPATHSEP!=tp[-1])
    *tp++=FFPATHSEP;
#endif // ]

  tree->temp.basename=tp;
#if defined (BG_TEMP_PREFIX) // [
  FFSTRCPY(tp,param->temp_prefix);
  tp+=len2;
#else // ] [
  *tp++=FFL('_');
  *tp++=FFL('_');
  *tp++=FFL('_');
  *tp++=FFL('_');
#endif // ]
  FFSTRCPY(tp,tree->target.basename);

  /////////////////////////////////////////////////////////////////////////////
  if (!ff_fcmp(tree->source.path,tree->temp.path)) {
    DMESSAGE("attempt to overwrite source file");
    goto e_overwrite_temp;
  }

#if defined (BG_TEMP_PREFIX) // [
  if (ff_fexists(tree->temp.path)) {
    FFVMESSAGE(FFL("file \"%s\" exists. Either remove it or by means of"
        " option --temp-prefix define a prefix for temporary files different"
        " from \"%s\""),
        tree->temp.path,param->temp_prefix);
    goto e_temp_exists;
  }
#endif // ]
#else // ] [
  /////////////////////////////////////////////////////////////////////////////
  if (!ff_fcmp(tree->source.path,tree->target.path)) {
    DMESSAGE("attempt to overwrite source file");
    goto e_overwrite_target;
  }
#endif // ]

#if defined (_WIN32) // [
  // if LANG is set to e.g. "en_US.UTF-8" we assume we're run from
  // e.g. MSYS2 shell undestanding UTF-8 otherwise from MS console using
  // codepage OEM. In the latter case we need an OEM representation of
  // the basename.
  if (tree->param->oem&&tree->target.basename) {
    ///////////////////////////////////////////////////////////////////////////
    track->target.oem.basename=bg_wcs2str(tree->target.basename,CP_OEMCP);

    if (!track->target.oem.basename) {
      DMESSAGE("creating oem basename");
      goto e_basename;
    }
  }
  else
    track->target.oem.basename=NULL;

  /////////////////////////////////////////////////////////////////////////////
#if defined (BG_WIN32_TARGET_UTF8) // [
  track->target.utf8.path=bg_wcs2str(tree->target.path,CP_UTF8);
  
  if (!track->target.utf8.path) {
    DMESSAGE("creating utf-8 representation of path");
    goto e_patha;
  }

  track->target.utf8.basename=bg_basenamea(track->target.utf8.path);
#else // ] [
  track->target.path=bg_wcs2str(tree->target.path,CP_UTF8);
  
  if (!track->target.path) {
    DMESSAGE("creating utf-8 representation of path");
    goto e_patha;
  }

  track->target.basename=bg_basenamea(track->target.path);
#endif // ]

#if defined (BG_MUXER_PATH_TEMP) // [
  /////////////////////////////////////////////////////////////////////////////
  track->temp.path=bg_wcs2str(tree->temp.path,CP_UTF8);
  
  if (!track->temp.path) {
    DMESSAGE("creating utf-8 representation of temporary path");
    goto e_tempa;
  }
#endif // ]
#endif // ]

  return 0;
//cleanup:
#if defined (_WIN32) // [
#if defined (BG_MUXER_PATH_TEMP) // [
  free(track->temp.path);
e_tempa:
#endif // ]
#if defined (BG_WIN32_TARGET_UTF8) // [
  free(track->target.utf8.path);
#else // ] [
  free(track->target.path);
#endif // ]
e_patha:
  if (track->target.oem.basename)
    free(track->target.oem.basename);
e_basename:
#endif // ]
#if defined (BG_TEMP_PREFIX) // [
e_temp_exists:
#endif // ]
e_overwrite_temp:
#if defined (BG_MUXER_PATH_TEMP) // [
  free(tree->temp.path);
e_temp:
#endif // ]
e_overwrite_target:
  free(tree->target.path);
e_path:
  return -1;
}

#if ! defined (BG_ANNOTATOR) // [
void bg_track_annotation_destroy(bg_tree_t *tree FFUNUSED)
{
#if defined (_WIN32) // [
  bg_track_t *track=&tree->track;

#if defined (BG_MUXER_PATH_TEMP) // [
  free(track->temp.path);
#endif // ]
#if defined (BG_WIN32_TARGET_UTF8) // [
  free(track->target.utf8.path);
#else // ] [
  free(track->target.path);
#endif // ]
  if (track->target.oem.basename)
    free(track->target.oem.basename);
#endif // ]
#if defined (BG_MUXER_PATH_TEMP) // [
  free(tree->temp.path);
#endif // ]
  free(tree->target.path);
}
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#if defined (BG_ANNOTATOR) // [
static int bg_annotator_dispatch_album(bg_visitor_t *vis, bg_tree_t *tree)
#else // ] [
int bg_album_annotation_create(bg_tree_t *tree)
#endif // ]
{
#if defined (BG_ANNOTATOR) // [
  bg_album_t *album=&tree->album;
#endif // ]
  bg_tree_t *parent=tree->parent;
  const ffchar_t *basename;
  size_t len1,len2,size;
  ffchar_t *tp;
#if defined (BG_ANNOTATOR) // [
  bg_tree_t *cur;
#endif // ]

  if (tree->param->output.dirname) {
    ///////////////////////////////////////////////////////////////////////////
    basename=bg_album_target_purge(tree);
    len1=parent&&parent->target.path?FFSTRLEN(parent->target.path):0u;
    len2=basename?FFSTRLEN(basename):0u;
    size=(len1?len1+1u:0u)+(len2?len2+1u:0u);

    ///////////////////////////////////////////////////////////////////////////
    if (0u<size) {
      /////////////////////////////////////////////////////////////////////////
      tree->target.path=tp=malloc(size*sizeof tp[0]);

      if (!tree->target.path) {
        DMESSAGE("allocating output path");
        goto e_path;
      }

      if (parent&&parent->target.path) {
        FFSTRCPY(tp,parent->target.path);
        tp+=len1;

        if (basename)
          *tp++=FFPATHSEP;
        else
          *tp=FFL('\0');
      }

      if (basename) {
        tree->target.basename=tp;
        FFSTRCPY(tp,basename);
        tp+=len2;
      }
      else
        tree->target.basename=NULL;
    }
    else {
      tree->target.basename=NULL;
      tree->target.path=NULL;
    }
  }
#if 0 && defined (BG_OVERWRITE) // [
  else if (tree->param->overwrite&&tree->source.path) {
    tree->target.path=FFSTRDUP(tree->source.path);

    if (!tree->target.path) {
      DMESSAGE("duplicating path");
      goto e_path;
    }

    tree->target.basename=bg_basename(tree->target.path);
  }
#endif // ]
  else {
    tree->target.basename=NULL;
    tree->target.path=NULL;
  }

#if defined (BG_ANNOTATOR) // [
  /////////////////////////////////////////////////////////////////////////////
  for (cur=album->first;cur;cur=cur->next) {
    if (cur->vtbl->accept(cur,vis)<0) {
      DMESSAGE("child accepting annotator");
      goto e_child_accept;
    }
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
#if defined (BG_ANNOTATOR) // [
e_child_accept:
#endif // ]
#if 0 // [
e_mkdir:
#endif // ]
  if (tree->target.path)
    free(tree->target.path);
e_path:
  return -1;
}

#if ! defined (BG_ANNOTATOR) // [
void bg_album_annotation_destroy(bg_tree_t *tree FFUNUSED)
{
  if (tree->target.path)
    free(tree->target.path);
}
#endif // ]


///////////////////////////////////////////////////////////////////////////////
#if defined (BG_ANNOTATOR) // [
static int bg_annotator_dispatch_root(bg_visitor_t *vis, bg_tree_t *tree)
#else // ] [
int bg_root_annotation_create(bg_tree_t *tree)
#endif // ]
{
#if defined (BG_ANNOTATOR) // [
  bg_tree_t *cur;
#endif // ]

  if (tree->param->output.dirname) {
    ///////////////////////////////////////////////////////////////////////////
    tree->target.path=FFSTRDUP(tree->param->output.dirname);

    if (!tree->target.path) {
      DMESSAGE("duplicating string");
      goto e_dup;
    }

#if 0 // [
    ///////////////////////////////////////////////////////////////////////////
    if (bg_mkdir(tree->target.path)<0) {
      DMESSAGE("creating directory");
      goto e_mkdir;
    }
#endif // ]
  }
  else
    tree->target.path=NULL;

#if defined (BG_ANNOTATOR) // [
  /////////////////////////////////////////////////////////////////////////////
  for (cur=tree->album.first;cur;cur=cur->next) {
    if (cur->vtbl->accept(cur,vis)<0) {
      DMESSAGE("accepting visitor");
      goto e_child;
    }
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  return 0;
// cleanup:
#if defined (BG_ANNOTATOR) // [
e_child:
#endif // ]
#if 0 // [
e_mkdir:
#endif // ]
  if (tree->target.path)
    free(tree->target.path);
e_dup:
  return -1;
}

#if ! defined (BG_ANNOTATOR) // [
void bg_root_annotation_destroy(bg_tree_t *tree FFUNUSED)
{
  if (tree->target.path)
    free(tree->target.path);
}
#endif // ]

#if defined (BG_ANNOTATOR) // [
///////////////////////////////////////////////////////////////////////////////
static bg_visitor_vtbl_t annotator_vtbl={
#if defined (PBU_DEBUG) // [
  .id="annotator",
#endif // ]
  .destroy=bg_annotator_destroy,
#if defined (BG_FILE) // [
  .dispatch_file=bg_annotator_dispatch_file,
#endif // ]
  .dispatch_track=bg_annotator_dispatch_track,
  .dispatch_album=bg_annotator_dispatch_album,
  .dispatch_root=bg_annotator_dispatch_root,
};
#endif // ]
