/*
 * bg_tree.c
 * Copyright (C) 2019 Peter Belkner <pbelkner@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <bg.h>

///////////////////////////////////////////////////////////////////////////////
int bg_tree_common_create(bg_tree_t *tree, bg_param_t *param,
    bg_tree_t *root, bg_tree_t *parent, const ffchar_t *path,
    bg_tree_vtbl_t *vtbl)
{
  /////////////////////////////////////////////////////////////////////////////
  memset(tree,0,sizeof *tree);

  /////////////////////////////////////////////////////////////////////////////
  tree->vtbl=vtbl;
  tree->param=param;
  tree->root=root;
  tree->parent=parent;
  tree->next=NULL;

  if (bg_tree_source_create(&tree->source,parent,path)<0) {
    DMESSAGE("creating source path");
    goto epath;
  }

#if defined (_WIN32) // [
  // if LANG is set to e.g. "en_US.UTF-8" we assume we're run from
  // e.g. MSYS2 shell undestanding UTF-8 otherwise from MS console using
  // codepage OEM. In the latter case we need an OEM representation of
  // the basename.
  if (param->oem&&tree->source.basename) {
    tree->oem.basename=bg_wcs2str(tree->source.basename,CP_OEMCP);

    if (!tree->oem.basename) {
      DMESSAGE("creating oem basename");
      goto ebasename;
    }
  }
  else
    tree->oem.basename=NULL;

  // in any case we need an utf-8 representation of path (end basaname.)
  if (bg_tree_patha_create(&tree->utf8,tree->source.path,CP_UTF8)<0) {
    DMESSAGE("creating patha");
    goto epatha;
  }
#endif // ]

#if defined (BG_LIB1770) // [
  /////////////////////////////////////////////////////////////////////////////
  if (BG_FLAGS_AGG_MOMENTARY&param->flags.aggregate) {
    tree->stats.momentary=lib1770_stats_new();

    if (!tree->stats.momentary) {
      DMESSAGE("creating momentary statistics");
      goto emomentary;
    }
  }
  else
    tree->stats.momentary=NULL;

  /////////////////////////////////////////////////////////////////////////////
  if (BG_FLAGS_AGG_SHORTTERM&param->flags.aggregate) {
    tree->stats.shortterm=lib1770_stats_new();

    if (!tree->stats.shortterm) {
      DMESSAGE("creating shortterm statistics");
      goto eshortterm;
    }
  }
  else
    tree->stats.shortterm=NULL;

  tree->stats.samplepeak=0.0;
  tree->stats.truepeak=0.0;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
#if defined (BG_LIB1770) // [
  if (tree->stats.shortterm)
    lib1770_stats_close(tree->stats.shortterm);
eshortterm:
  if (tree->stats.momentary)
    lib1770_stats_close(tree->stats.momentary);
emomentary:
#endif // ]
#if defined (_WIN32) // [
  bg_tree_patha_destroy(&tree->utf8);
epatha:
  if (tree->oem.basename)
    free(tree->oem.basename);
ebasename:
#endif // ]
  bg_tree_path_destroy(&tree->source);
epath:
  return -1;
}

void bg_tree_common_destroy(bg_tree_t *tree)
{
#if defined (BG_ANNOTATOR) // [
#if defined (BG_MUXER_PATH_TEMP) // [
  if (tree->temp.path)
    free(tree->temp.path);
#endif // ]

  if (tree->target.path)
    free(tree->target.path);
#endif // ]

#if defined (BG_LIB1770) // [
  if (tree->stats.shortterm)
    lib1770_stats_close(tree->stats.shortterm);

  if (tree->stats.momentary)
    lib1770_stats_close(tree->stats.momentary);
#endif // ]

#if defined (_WIN32) // [
  bg_tree_patha_destroy(&tree->utf8);

  if (tree->oem.basename)
    free(tree->oem.basename);
#endif // ]

  bg_tree_path_destroy(&tree->source);
}

#if defined (BG_TYPE_DIR) // [
int bg_child_create(bg_tree_t **tree, bg_param_t *param, bg_tree_t *root,
    bg_tree_t *parent, const ffchar_t *path, int dir)
#else // ] [
int bg_child_create(bg_tree_t **tree, bg_param_t *param, bg_tree_t *root,
    bg_tree_t *parent, const ffchar_t *path)
#endif // ]
{
  /////////////////////////////////////////////////////////////////////////////
  *tree=malloc(sizeof **tree);

  if (!*tree) {
#if defined (BG_TREE_CREATE_CHILD_WARNING) // [
    DWARNING("allocating tree");
#endif // ]
    goto e_malloc;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (bg_tree_common_create(*tree,param,root,parent,path,NULL)<0) {
#if defined (BG_TREE_CREATE_CHILD_WARNING) // [
    DWARNING("creating tree");
#endif // ]
    goto e_tree;
  }

  /////////////////////////////////////////////////////////////////////////////
#if defined (BG_TYPE_DIR) // [
  if (dir) {
    if (bg_album_create(*tree,param)<0) {
      DMESSAGE("creating album");
      goto e_album;
    }
  }
#else // ] [
  if (0<=bg_album_create(*tree,param))
    ;
#endif // ]
  else if (0<=bg_track_create(*tree))
    ;
#if defined (BG_FILE) // [
#if defined (BG_OVERWRITE) // [
  else if (!param->overwrite
      &&(BG_FLAGS_EXT_COPY&param->flags.extension)
      &&0<=bg_file_create(*tree))
    ;
#else // ] [
  else if ((BG_FLAGS_EXT_COPY&param->flags.extension)
      &&0<=bg_file_create(*tree))
    ;
#endif // ]
#endif // ]
  else {
#if defined (BG_TREE_CREATE_CHILD_WARNING) // [
    FFVWARNING(FFL("creating child \"%s\""),(*tree)->path.source);
#endif // ]
    goto e_child;
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
  (*tree)->vtbl->destroy(*tree);
e_child:
  bg_tree_common_destroy(*tree);
#if defined (BG_TYPE_DIR) // [
e_album:
#endif // ]
e_tree:
  free(*tree);
  *tree=NULL;
e_malloc:
  return -1;
}

void bg_child_destroy(bg_tree_t **tree)
{
  (*tree)->vtbl->destroy(*tree);
  bg_tree_common_destroy(*tree);
  free(*tree);
  *tree=NULL;
}

#if defined (BG_LIB1770) // [
double *bg_tree_samplepeak(bg_tree_t *tree)
{
  return tree->param->flags.aggregate&BG_FLAGS_AGG_SAMPLEPEAK
      ?&tree->stats.samplepeak:NULL;
}

double *bg_tree_truepeak(bg_tree_t *tree)
{
  return tree->param->flags.aggregate&BG_FLAGS_AGG_TRUEPEAK
      ?&tree->stats.truepeak:NULL;
}

int bg_tree_merge(bg_tree_t *lhs, const bg_tree_t *rhs)
{
  bg_flags_agg_t aggregate=lhs->param->flags.aggregate;

  if (aggregate!=rhs->param->flags.aggregate) {
    DMESSAGE("trying to merge incompatible trees");
    goto emerge;
  }

  if (0!=(BG_FLAGS_AGG_MOMENTARY&aggregate))
    lib1770_stats_merge(lhs->stats.momentary,rhs->stats.momentary);

  if (0!=(BG_FLAGS_AGG_SHORTTERM&aggregate))
    lib1770_stats_merge(lhs->stats.shortterm,rhs->stats.shortterm);

  if (0!=(BG_FLAGS_AGG_SAMPLEPEAK&aggregate)) {
    if (lhs->stats.samplepeak<rhs->stats.samplepeak)
      lhs->stats.samplepeak=rhs->stats.samplepeak;
  }

  if (0!=(BG_FLAGS_AGG_TRUEPEAK&aggregate)) {
    if (lhs->stats.truepeak<rhs->stats.truepeak)
      lhs->stats.truepeak=rhs->stats.truepeak;
  }

  return 0;
emerge:
  return -1;
}

#if defined (_WIN32) // [
const char *bg_tree_in_basanamen(bg_tree_t *tree)
{
  return tree->oem.basename?tree->oem.basename:tree->utf8.basename;
}

const wchar_t *bg_tree_in_basanamew(bg_tree_t *tree)
{
  return tree->source.basename;
}

const char *bg_tree_out_basanamen(bg_tree_t *tree)
{
#if 0 // [
  // TODO: source.utf8 -> target.oem
  return tree->oem.basename?tree->oem.basename:tree->utf8.basename;
#else // ] [
  bg_track_t *track=&tree->track;

  // just needed for tracks.
  if (BG_TREE_TYPE_TRACK!=tree->vtbl->type) {
    DVMESSAGE("unexpected tree type %d",tree->vtbl->type);
    return NULL;
  }
  else {
#if defined (BG_WIN32_TARGET_UTF8) // [
    return track->target.oem.basename
        ?track->target.oem.basename:track->target.utf8.basename;
#else // ] [
    return track->target.oem.basename
        ?track->target.oem.basename:track->target.basename;
#endif // ]
  }
#endif // ]
}

const wchar_t *bg_tree_out_basanamew(bg_tree_t *tree)
{
  return tree->target.basename;
}
#else // ] [
const char *bg_tree_in_basanamen(bg_tree_t *tree)
{
  return tree->source.basename;
}

const char *bg_tree_out_basanamen(bg_tree_t *tree)
{
  return tree->target.basename;
}
#endif // ]

#if ! defined (BG_PRINT) // [
void bg_tree_result(const bg_tree_t *tree, FILE *f)
{
  bg_param_t *param=tree->param;
  double norm=param->preamp+param->norm;
  double db,q;

  // momentary ///////////////////////////////////////////////////////////////
  if (tree->param->flags.aggregate&BG_FLAGS_AGG_MOMENTARY_MEAN) {
    db=lib1770_stats_get_mean(tree->stats.momentary,
        param->momentary.mean.gate);
#if defined (_WIN32) // [
    if (stdout==f||stderr==f) {
      // If writing to the console we need to write narrow otherwise wide
      // charcters.
      fprintf(f,"    integrated (momentary mean): %.2f LUFS / %.2f LU\n",
          db,norm-db);
    }
    else {
      fwprintf(f,L"    integrated (momentary mean): %.2f LUFS / %.2f LU\n",
          db,norm-db);
    }
#else // ] [
    fprintf(f,"    integrated (momentary mean): %.2f LUFS / %.2f LU\n",
        db,norm-db);
#endif // ]
  }

  if (tree->param->flags.aggregate&BG_FLAGS_AGG_MOMENTARY_MAXIMUM) {
    db=lib1770_stats_get_max(tree->stats.momentary);
#if defined (_WIN32) // [
    if (stdout==f||stderr==f) {
      // If writing to the console we need to write narrow otherwise wide
      // charcters.
      fprintf(f,"    momentary maximum: %.2f LUFS / %.2f LU\n",
          db,norm-db);
    }
    else {
      fwprintf(f,L"    momentary maximum: %.2f LUFS / %.2f LU\n",
          db,norm-db);
    }
#else // ] [
    fprintf(f,"    momentary maximum: %.2f LUFS / %.2f LU\n",
        db,norm-db);
#endif // ]
  }

  if (tree->param->flags.aggregate&BG_FLAGS_AGG_MOMENTARY_RANGE) {
    db=lib1770_stats_get_range(tree->stats.momentary,
        param->momentary.range.gate,
        param->momentary.range.lower_bound,
        param->momentary.range.upper_bound);
#if defined (_WIN32) // [
    if (stdout==f||stderr==f) {
      // If writing to the console we need to write narrow otherwise wide
      // charcters.
      fprintf(f,"    momentary range: %.2f LRA\n",db);
    }
    else
      fwprintf(f,L"    momentary range: %.2f LRA\n",db);
#else // ] [
    fprintf(f,"    momentary range: %.2f LRA\n",db);
#endif // ]
  }

  // shortterm ///////////////////////////////////////////////////////////////
  if (tree->param->flags.aggregate&BG_FLAGS_AGG_SHORTTERM_MEAN) {
    db=lib1770_stats_get_mean(tree->stats.shortterm,
        param->momentary.mean.gate);
#if defined (_WIN32) // [
    if (stdout==f||stderr==f) {
      // If writing to the console we need to write narrow otherwise wide
      // charcters.
      fprintf(f,"    shortterm mean: %.2f LUFS / %.2f LU\n",
          db,norm-db);
    }
    else {
      fwprintf(f,L"    shortterm mean: %.2f LUFS / %.2f LU\n",
          db,norm-db);
    }
#else // ] [
    fprintf(f,"    shortterm mean: %.2f LUFS / %.2f LU\n",
        db,norm-db);
#endif // ]
  }

  if (tree->param->flags.aggregate&BG_FLAGS_AGG_SHORTTERM_MAXIMUM) {
    db=lib1770_stats_get_max(tree->stats.shortterm);
#if defined (_WIN32) // [
    if (stdout==f||stderr==f) {
      // If writing to the console we need to write narrow otherwise wide
      // charcters.
      fprintf(f,"    shortterm maximum: %.2f LUFS / %.2f LU\n",
          db,norm-db);
    }
    else {
      fwprintf(f,L"    shortterm maximum: %.2f LUFS / %.2f LU\n",
          db,norm-db);
    }
#else // ] [
    fprintf(f,"    shortterm maximum: %.2f LUFS / %.2f LU\n",
        db,norm-db);
#endif // ]
  }

  if (tree->param->flags.aggregate&BG_FLAGS_AGG_SHORTTERM_RANGE) {
    db=lib1770_stats_get_range(tree->stats.shortterm,
        param->shortterm.range.gate,
        param->shortterm.range.lower_bound,
        param->shortterm.range.upper_bound);
#if defined (_WIN32) // [
    if (stdout==f||stderr==f) {
      // If writing to the console we need to write narrow otherwise wide
      // charcters.
      fprintf(f,"    shortterm range: %.2f LRA\n",db);
    }
    else
      fwprintf(f,L"    shortterm range: %.2f LRA\n",db);
#else // ] [
    fprintf(f,"    shortterm range: %.2f LRA\n",db);
#endif // ]
  }

  // samplepeak //////////////////////////////////////////////////////////////
  if (tree->param->flags.aggregate&BG_FLAGS_AGG_SAMPLEPEAK) {
    q=tree->stats.samplepeak;

    if (q) {
#if defined (_WIN32) // [
      if (stdout==f||stderr==f) {
        // If writing to the console we need to write narrow otherwise wide
        // charcters.
        fprintf(f,"    sample peak: %.2f SPFS / %f\n",
            LIB1770_Q2DB(q),q);
      }
      else {
        fwprintf(f,L"    sample peak: %.2f SPFS / %f\n",
            LIB1770_Q2DB(q),q);
      }
#else // ] [
      fprintf(f,"    sample peak: %.2f SPFS / %f\n",
          LIB1770_Q2DB(q),q);
#endif // ]
    }
    else {
#if defined (_WIN32) // [
      if (stdout==f||stderr==f) {
        // If writing to the console we need to write narrow otherwise wide
        // charcters.
        fprintf(f,"    sample peak: %f\n",q);
      }
      else
        fwprintf(f,L"    sample peak: %f\n",q);
#else // ] [
      fprintf(f,"    sample peak: %f\n",q);
#endif // ]
    }
  }

  // truepeak ////////////////////////////////////////////////////////////////
  if (tree->param->flags.aggregate&BG_FLAGS_AGG_TRUEPEAK) {
    q=tree->stats.truepeak;

    if (q) {
#if defined (_WIN32) // [
      if (stdout==f||stderr==f) {
        // If writing to the console we need to write narrow otherwise wide
        // charcters.
        fprintf(f,"    true peak: %.2f TPFS / %f\n",
            LIB1770_Q2DB(q),q);
      }
      else {
        fwprintf(f,L"    true peak: %.2f TPFS / %f\n",
            LIB1770_Q2DB(q),q);
      }
#else // ] [
      fprintf(f,"    true peak: %.2f TPFS / %f\n",
          LIB1770_Q2DB(q),q);
#endif // ]
    }
    else {
#if defined (_WIN32) // [
      if (stdout==f||stderr==f) {
        // If writing to the console we need to write narrow otherwise wide
        // charcters.
        fprintf(f,"    true peak: %f\n",q);
      }
      else
        fwprintf(f,L"    true peak: %f\n",q);
#else // ] [
      fprintf(f,"    true peak: %f\n",q);
#endif // ]
    }
  }
}
#endif // ]
#endif // ]
