/*
 * ffm.c
 * Copyright (C) 2019 Peter Belkner <pbelkner@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
/*
 * an example program on the implementation of libff.a.
 */
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <getopt.h>
#include <libavutil/opt.h>
#include <libswresample/swresample.h>
#include <libavformat/avformat.h>
#include <libavfilter/buffersink.h>
#include <libavfilter/buffersrc.h>
#include <lib1770.h>

///////////////////////////////////////////////////////////////////////////////
#ifdef __GNUC__ // [
#define UNUSED __attribute__((__unused__))
#else // ] [
#define UNUSED
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#define DMESSAGE(format) { \
  av_log(NULL,AV_LOG_ERROR,"Error: " format ". (%s:%d:%s)\n", \
      basename(__FILE__),__LINE__,__func__); \
  fflush(stderr); \
}
#define DVMESSAGE(format, ...) { \
  av_log(NULL,AV_LOG_ERROR,"Error: " format ". (%s:%d:%s)\n", \
      __VA_ARGS__,basename(__FILE__),__LINE__,__func__); \
  fflush(stderr); \
}
#define DWARNING(format) { \
  av_log(NULL,AV_LOG_ERROR,"Warning: " format ". (%s:%d:%s)\n", \
      basename(__FILE__),__LINE__,__func__); \
  fflush(stderr); \
}
#define DVWARNING(format, ...) { \
  av_log(NULL,AV_LOG_ERROR,"Warning: " format ". (%s:%d:%s)\n", \
      __VA_ARGS__,basename(__FILE__),__LINE__,__func__); \
  fflush(stderr); \
}
#define DMARKLN() { \
  av_log(NULL,AV_LOG_INFO,"%s:%d:%s\n", \
      basename(__FILE__),__LINE__,__func__); \
  fflush(stderr); \
}
#define DWRITELN(format) { \
  av_log(NULL,AV_LOG_INFO,format " (%s:%d:%s)\n", \
      basename(__FILE__),__LINE__,__func__); \
  fflush(stderr); \
}
#define DVWRITELN(format, ...) { \
  av_log(NULL,AV_LOG_INFO,format " (%s:%d:%s)\n", \
    __VA_ARGS__,basename(__FILE__),__LINE__,__func__); \
  fflush(stderr); \
}

///////////////////////////////////////////////////////////////////////////////
//#define MUXER_SAMPLE_FORMAT   AV_SAMPLE_FMT_S32
#define FIX_CHANNEL_LAYOUT
#define PACKET_UNREF
#define FRAME_UNREF

///////////////////////////////////////////////////////////////////////////////
typedef struct {
  struct {
	  enum AVSampleFormat sample_fmt;
    int64_t channel_layout;
  } request;

  struct {
    int enabled;
    double scale;
  } drc;
} param_decode_t;

////////
typedef struct {
  AVCodecContext *ctx;
} audio_t;

////////
typedef struct {
  struct {
    AVFormatContext *ctx;
  } fmt;

  int ai,vi;
  audio_t audio;
} inout_t;

////////
typedef struct {
  SwrContext *ctx;
  AVFrame *frame;
} resampler_t;

////////
typedef const struct {
  const param_decode_t *(*decode)(const void *data);

  struct {
    int (*create)(void *data, const AVCodecParameters *codecpar);
    void (*destroy)(void *data);
    int (*add)(void *data, int upsampled, AVFrame *frame);
  } stats;
} analyzer_callback_t;

typedef enum {
  ANALYZER_DECODER_SEND_PACKET,
  ANALYZER_DECODER_RECEIVE_FRAME,
} analyzer_state_t;

////////
typedef struct {
  analyzer_state_t state;
  void *data;
  analyzer_callback_t *cb;
  inout_t *in;
  resampler_t normalizer;
  resampler_t upsampler;
  AVPacket *pkt;
  AVFrame *frame;
} analyzer_t;

////////
typedef struct {
  struct {
    AVFilterContext *sink,*src;
  } ctx;

  AVFilterGraph *graph;
} filter_t;

////////
typedef enum {
  MUXER_DECODER_SEND_PACKET,
  MUXER_DECODER_RECEIVE_FRAME,
  MUXER_FILTER_SEND_FRAME,
  MUXER_FILTER_RECEIVE_FRAME,
  MUXER_ENCODER_SEND_FRAME,
  MUXER_ENCODER_RECEIVE_PACKET,
} muxer_state_t;

typedef struct {
  muxer_state_t state;
  inout_t *in,*out;
  AVPacket *pkt;
  filter_t filter;
  AVFrame *frame;
} muxer_t;

////////
typedef const struct iter_vmt iter_vmt_t;
typedef struct iter iter_t;

struct iter_vmt {
#if defined (PBU_DEBUG) // [
  const char *id;
#endif // ]
  void (*first)(iter_t *i);
  int (*valid)(iter_t *i);
  void (*next)(iter_t *i);
  void (*norm)(iter_t *i, double *x, double *max);
};

struct iter {
  iter_vmt_t *vmt;
  const AVFrame *frame;

	union {
    // interleaved ////////////////////////////////////////////////////////////
    union {
      union {
        struct { int16_t *rp,*mp; } s16;
        struct { int32_t *rp,*mp; } s32;
        struct { float   *rp,*mp; } flt;
        struct { double  *rp,*mp; } dbl;
      };
    } i;

    // planar /////////////////////////////////////////////////////////////////
    union {
      union {
        struct { int16_t *rp[AV_NUM_DATA_POINTERS],*mp; } s16;
        struct { int32_t *rp[AV_NUM_DATA_POINTERS],*mp; } s32;
        struct { float   *rp[AV_NUM_DATA_POINTERS],*mp; } flt;
        struct { double  *rp[AV_NUM_DATA_POINTERS],*mp; } dbl;
      };
    } p;
  };
};

///////////////////////////////////////////////////////////////////////////////
char *basenamex(char *path, char *seps)
{
  char *pp,*sp;

  for (pp=path+strlen(path);path<pp;--pp) {
    for (sp=seps;*sp;++sp) {
      if (*sp==pp[-1])
        return pp;
    }
  }

  return pp;
}

char *basename(char *path)
{
  return basenamex(path,"/\\");
}

///////////////////////////////////////////////////////////////////////////////
int audio_create(audio_t *audio, inout_t *inout, const param_decode_t *iparam,
    const AVCodecParameters *ocodecpar)
{
  AVStream *stream=inout->fmt.ctx->streams[inout->ai];
  AVCodecParameters *codecpar=stream->codecpar;
  AVDictionary *opt=NULL;
  AVCodec *codec;
  int i;
  char value[64];
  int err;

  ////////////////////////////////////////////////////////////////////////////
	if ((iparam&&ocodecpar)||(!iparam&&!ocodecpar)) {
		DMESSAGE("unexpected arguments");
		goto eargs;
	}

  ////////////////////////////////////////////////////////////////////////////
  if (ocodecpar) {
    // we're going to encode and will construct the corresponding
    // stream's codec parameters below.
    codec=avcodec_find_encoder(ocodecpar->codec_id);

    if (!codec) {
      DVMESSAGE("audio encoder \"%s\" not available",
          avcodec_get_name(ocodecpar->codec_id));
      goto ecodec;
    }
  }
  else {
    // we're going to decode and receive needed information from the
    // correponding stream's codec parameters.
    codec=avcodec_find_decoder(codecpar->codec_id);

    if (!codec) {
      DVMESSAGE("audio decoder \"%s\" not available",
          avcodec_get_name(codecpar->codec_id));
      goto ecodec;
    }
  }

  ////////////////////////////////////////////////////////////////////////////
  audio->ctx=avcodec_alloc_context3(codec);

  if (!audio->ctx) {
    DMESSAGE("allocating codec context");
    goto econtext;
  }

  ////////////////////////////////////////////////////////////////////////////
  if (ocodecpar) {
    // we're going to encode.
    if (codec->sample_fmts) {
#if 0 // [
      audio->ctx->sample_fmt=codec->sample_fmts[0];
#else // ] [
      audio->ctx->sample_fmt=AV_SAMPLE_FMT_NONE;
#endif // ]

      for (i=0;0<=codec->sample_fmts[i];++i) {
        if (codec->sample_fmts[i]==ocodecpar->format) {
          audio->ctx->sample_fmt=ocodecpar->format;
          break;
        }
      }

      if (AV_SAMPLE_FMT_NONE==audio->ctx->sample_fmt) {
        DMESSAGE("requested sample format not supported by the container");
        goto ecopy1;
      }
    }
    else
      audio->ctx->sample_fmt=ocodecpar->format;

    if (codec->supported_samplerates) {
      audio->ctx->sample_rate=codec->supported_samplerates[0];

      for (i=0;codec->supported_samplerates[i];++i) {
        if (codec->supported_samplerates[i]==ocodecpar->sample_rate) {
          audio->ctx->sample_rate=ocodecpar->sample_rate;
          break;
        }
      }
    }
    else
      audio->ctx->sample_rate=ocodecpar->sample_rate;

    if (codec->channel_layouts) {
      audio->ctx->channel_layout=codec->channel_layouts[0];

      for (i=0;codec->channel_layouts[i];++i) {
        if (codec->channel_layouts[i]==ocodecpar->channel_layout) {
          audio->ctx->channel_layout=ocodecpar->channel_layout;
          break;
        }
      }
    }
    else
      audio->ctx->channel_layout=ocodecpar->channel_layout;

    audio->ctx->channels
        =av_get_channel_layout_nb_channels(audio->ctx->channel_layout);
    audio->ctx->time_base.num=1;
    audio->ctx->time_base.den=audio->ctx->sample_rate;
    stream->time_base=audio->ctx->time_base;
  }
  else if (iparam) {
    // we're going to decode.
    err=avcodec_parameters_to_context(audio->ctx,codecpar);

    if (err<0) {
      DVMESSAGE("copying codec parameters: %s (%d)",av_err2str(err),err);
      goto ecopy1;
    }

    if (0ll<=iparam->request.channel_layout)
      audio->ctx->request_channel_layout=iparam->request.channel_layout;

		if (0<=iparam->request.sample_fmt)
    	audio->ctx->request_sample_fmt=iparam->request.sample_fmt;

    ///////////////////////////////////////////////////////////////////////////
    if (iparam->drc.enabled&&AV_CODEC_ID_AC3==codecpar->codec_id) {
      sprintf(value,"%0.1f",iparam->drc.scale);
      err=av_dict_set(&opt,"drc_scale",value,0);

      if (err<0) {
        DVMESSAGE("setting drc_scale: %s (%d)",av_err2str(err),err);
        goto ecopy1;
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  err=avcodec_open2(audio->ctx,codec,&opt);
  av_dict_free(&opt);

  if (err<0) {
    DVMESSAGE("copening codec context: %s (%d)",av_err2str(err),err);
    goto eopen;
  }

#if defined (FIX_CHANNEL_LAYOUT) // [
  if (iparam&&!audio->ctx->channel_layout) {
    // this is some strange behaviour: when we request AV_CH_LAYOUT_STEREO
    // from AV_CH_LAYOUT_5POINT1 it might turn out that everything is fine
    // (especially "channels") except "channel_layout" might not be set.
    audio->ctx->channel_layout=iparam->request.channel_layout;
  }
#endif // ]

#if 0 // [
  if (codecpar) {
    // we're going to encode and as promised above we copy the codec
    // context's parameters to the corresponding stream.
#else // ] [
    // we copy the codec parameters anyway because "sample_fmt" might
    // have changed in case of decoding.
#endif // ]
    err=avcodec_parameters_from_context(codecpar,audio->ctx);

    if (err<0) {
      DVMESSAGE("copying codec parameters: %s (%d)",av_err2str(err),err);
      goto ecopy2;
    }
#if 0 // [
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
ecopy2:
eopen:
ecopy1:
  avcodec_free_context(&audio->ctx);
econtext:
ecodec:
eargs:
  return -1;
}

void audio_destroy(audio_t *audio)
{
  avcodec_free_context(&audio->ctx);
}

///////////////////////////////////////////////////////////////////////////////
int input_create(inout_t *in, const char *path, const param_decode_t *param)
{
  int i;
  const AVCodecParameters *codecpar;
  int err;

  /////////////////////////////////////////////////////////////////////////////
  in->fmt.ctx=NULL;
  err=avformat_open_input(&in->fmt.ctx,path,NULL,NULL);

  if (err<0) {
    DVMESSAGE("opening input: %s (%d)",av_err2str(err),err);
    goto econtext;
  }

  /////////////////////////////////////////////////////////////////////////////
  // should be set before calling avformat_find_stream_info().
  // cf. e.g. https://code.videolan.org/libav/libav/commit/9dc0bc3dadbc01b350b84c0079de391cb6015094
  in->fmt.ctx->flags|=AVFMT_FLAG_GENPTS;
  err=avformat_find_stream_info(in->fmt.ctx,NULL);

  if (err<0) {
    DVMESSAGE("finding stream info: %s (%d)",av_err2str(err),err);
    goto efind;
  }

  /////////////////////////////////////////////////////////////////////////////
  av_dump_format(in->fmt.ctx,0,path,0);
  fflush(stderr);

  /////////////////////////////////////////////////////////////////////////////
  in->ai=-1;
  in->vi=-1;

  for (i=0;i<(int)in->fmt.ctx->nb_streams;++i) {
    codecpar=in->fmt.ctx->streams[i]->codecpar;

    if (codecpar->codec_id<=AV_CODEC_ID_NONE)
      continue;
    else if (AV_CODEC_ID_FIRST_SUBTITLE<=codecpar->codec_id)
      continue;

    switch (codecpar->codec_type) {
    case AVMEDIA_TYPE_AUDIO:
      if (codecpar->channels<=0)
        continue;
      else if (AV_NUM_DATA_POINTERS<=codecpar->channels)
        continue;
      else if (in->ai<0) {
        // if no audio stream found yet, pick this one (regardless whether
        // stereo or not).
        in->ai=i;
      }
      else if (2==codecpar->channels) {
        // if an audio stream alreay was found (and is non-stereo),
        // replace it by this stereo one.
        in->ai=i;
      }

      break;
    case AVMEDIA_TYPE_VIDEO:
      if (in->vi<0) {
        // if no video stream found yet, pick this one.
        in->vi=i;
      }

      break;
    default:
      break;
    }
  }

  if (in->ai<0) {
    DMESSAGE("missing audio stream");
    goto enoaudio;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (!param)
    in->audio.ctx=NULL;
  else if (audio_create(&in->audio,in,param,NULL)<0) {
    DMESSAGE("creating audio decoder");
    goto eaudio;
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
  if (in->audio.ctx)
    audio_destroy(&in->audio);
eaudio:
enoaudio:
efind:
  avformat_close_input(&in->fmt.ctx);
econtext:
  return -1;
}

void input_destroy(inout_t *in)
{
  if (in->audio.ctx)
    audio_destroy(&in->audio);

  avformat_close_input(&in->fmt.ctx);
}

int input_open(inout_t *in, const char *path, const param_decode_t *param)
{
  int err;

  /////////////////////////////////////////////////////////////////////////////
  if (in->ai<0) {
    DMESSAGE("audio index out of range");
    goto erange;
  }

  /////////////////////////////////////////////////////////////////////////////
  in->fmt.ctx=NULL;
  err=avformat_open_input(&in->fmt.ctx,path,NULL,NULL);

  if (err<0) {
    DVMESSAGE("opening input: %s (%d)",av_err2str(err),err);
    goto econtext;
  }

  /////////////////////////////////////////////////////////////////////////////
  // should be set before calling avformat_find_stream_info().
  // cf. e.g. https://code.videolan.org/libav/libav/commit/9dc0bc3dadbc01b350b84c0079de391cb6015094
  in->fmt.ctx->flags|=AVFMT_FLAG_GENPTS;
  err=avformat_find_stream_info(in->fmt.ctx,NULL);

  if (err<0) {
    DVMESSAGE("finding stream info: %s (%d)",av_err2str(err),err);
    goto efind;
  }

#if 0 // [
  /////////////////////////////////////////////////////////////////////////////
  av_dump_format(in->fmt.ctx,0,path,0);
  fflush(stderr);
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  if (!param)
    in->audio.ctx=NULL;
  else if (audio_create(&in->audio,in,param,NULL)<0) {
    DMESSAGE("creating audio decoder");
    goto eaudio;
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
  if (in->audio.ctx)
    audio_destroy(&in->audio);
eaudio:
efind:
  avformat_close_input(&in->fmt.ctx);
econtext:
erange:
  return -1;
}

void input_close(inout_t *in)
{
  if (in->audio.ctx)
    audio_destroy(&in->audio);

  avformat_close_input(&in->fmt.ctx);
}

#if 0 // [
const AVCodecParameters *input_codecpar(const inout_t *in)
{
  return in->fmt.ctx->streams[in->ai]->codecpar;
}
#endif // ]

void input_progress(inout_t *in, AVPacket *pkt, FILE *f)
{
  AVStream *stream=in->fmt.ctx->streams[pkt->stream_index];
  int64_t duration=av_rescale_q(in->fmt.ctx->duration,AV_TIME_BASE_Q,
      stream->time_base);
  double percent=0ll<pkt->pts&&0ll<duration?100.0*pkt->pts/duration:0.0;
  char buf[32];
  int len;

  len=sprintf(buf,"%.0f%%",percent);
  fputs(buf,f);

  while (len) {
    fputc('\b',f);
    --len;
  }

  fflush(f);
}

///////////////////////////////////////////////////////////////////////////////
int output_create(inout_t *out, const char *path, const analyzer_t *a,
    enum AVSampleFormat sample_fmt)
{
  const param_decode_t *decode
      =a->cb&&a->cb->decode?a->cb->decode(a->data):NULL;
  int64_t channel_layout=decode?decode->request.channel_layout:-1ll;
  inout_t *in=a->in;
  int err;
  AVStream *istream,*ostream;
  AVCodecParameters codecpar;

  /////////////////////////////////////////////////////////////////////////////
  out->fmt.ctx=NULL;
  err=avformat_alloc_output_context2(&out->fmt.ctx,NULL,NULL,path);

  if (err<0) {
    DVMESSAGE("allocating output context: %s (%d)",av_err2str(err),err);
    goto eformat;
  }


  /////////////////////////////////////////////////////////////////////////////
  if (0<=in->vi) {
    istream=in->fmt.ctx->streams[in->vi];

		// err is interpreted as a boolean.
    if (out->fmt.ctx->oformat->query_codec)
      err=out->fmt.ctx->oformat->query_codec(istream->codecpar->codec_id,1);
    else
      err=0;

    if (err) {
      istream=in->fmt.ctx->streams[in->vi];
      out->vi=out->fmt.ctx->nb_streams;

      // if succeeding the stream will be added to out->fmt.ctx and hence
      // will be destroyed by calling avformat_free_context().
      ostream=avformat_new_stream(out->fmt.ctx,NULL);

      if (!ostream) {
        DVMESSAGE("creating video output stream: %s (%d)",av_err2str(err),err);
        goto evideo;
      }

      // we're going to remux.
      err=avcodec_parameters_copy(ostream->codecpar,istream->codecpar);

      if (err<0) {
        DMESSAGE("copying codec parametrs");
        goto evideo;
      }

      ostream->id=out->vi;
      ostream->time_base=istream->time_base;
      ostream->codecpar->codec_tag=0;
    }
    else {
      // the video is lost.
			DWARNING("unable to remux video stream");
      in->vi=-1;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  if (0<=in->ai) {
    istream=in->fmt.ctx->streams[in->ai];
    out->ai=out->fmt.ctx->nb_streams;

    // if succeeding the stream will be added to out->fmt.ctx and hence
    // will be destroyed by calling avformat_free_context().
    ostream=avformat_new_stream(out->fmt.ctx,NULL);

    if (!ostream) {
      DVMESSAGE("creating audio output stream: %s (%d)",av_err2str(err),err);
      goto eaudio;
    }

    ostream->id=out->ai;

    if (!in->audio.ctx) {
      out->audio.ctx=NULL;
      err=avcodec_parameters_copy(ostream->codecpar,istream->codecpar);

      if (err<0) {
        DMESSAGE("copying codec parametrs");
        goto eaudio;
      }

      ostream->time_base=istream->time_base;
      ostream->codecpar->codec_tag=0;
    }
    else {
      codecpar=*istream->codecpar;
      codecpar.codec_id=AV_CODEC_ID_FLAC;
      codecpar.format=sample_fmt;

      if (0ll<=channel_layout)
        codecpar.channel_layout=channel_layout;

      err=audio_create(&out->audio,out,NULL,&codecpar);
      fflush(stderr);

      if (err<0) {
        DVMESSAGE("creating audio encoder: %s (%d)",av_err2str(err),err);
        goto eaudio;
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  av_dump_format(out->fmt.ctx,0,path,1);
  fflush(stderr);

  /////////////////////////////////////////////////////////////////////////////
  if (!(out->fmt.ctx->oformat->flags&AVFMT_NOFILE)) {
    err=avio_open(&out->fmt.ctx->pb,path,AVIO_FLAG_WRITE);

    if (err<0) {
      DVMESSAGE("open input file: %s (%d)",av_err2str(err),err);
      goto eopen;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  err=avformat_write_header(out->fmt.ctx,NULL);

  if (err<0) {
    DVMESSAGE("writing header: %s (%d)",av_err2str(err),err);
    goto eheader;
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
  av_write_trailer(out->fmt.ctx);
eheader:
  if (!(out->fmt.ctx->oformat->flags&AVFMT_NOFILE))
    avio_closep(&out->fmt.ctx->pb);
eopen:
  if (out->audio.ctx)
    audio_destroy(&out->audio);
eaudio:
evideo:
  avformat_free_context(out->fmt.ctx);
eformat:
  return -1;
}

void output_destroy(inout_t *out)
{
  av_write_trailer(out->fmt.ctx);

  if (!(out->fmt.ctx->oformat->flags&AVFMT_NOFILE))
    avio_closep(&out->fmt.ctx->pb);

  if (out->audio.ctx)
    audio_destroy(&out->audio);

  avformat_free_context(out->fmt.ctx);
}

///////////////////////////////////////////////////////////////////////////////
int resampler_create(resampler_t *res,
    const AVCodecParameters *ocodecpar,
    const AVCodecParameters *icodecpar)
{
  int64_t ochannel_layout=ocodecpar->channel_layout;

  /////////////////////////////////////////////////////////////////////////////
  res->ctx=swr_alloc_set_opts(
      NULL,                       // struct SwrContext *s,
      // [
      ochannel_layout,            // int64_t out_ch_layout,
      ocodecpar->format,          // enum AVSampleFormat out_sample_fmt,
      ocodecpar->sample_rate,     // int out_sample_rate,
      // ] [
      icodecpar->channel_layout,  // int64_t in_ch_layout,
      icodecpar->format,          // enum AVSampleFormat in_sample_fmt,
      icodecpar->sample_rate,     // int in_sample_rate,
      // ]
      0,                          // int log_obgset,
      NULL);                      // void *log_ctx

  if (!res->ctx) {
    DMESSAGE("allocating context");
    goto econtext;
  }

  /////////////////////////////////////////////////////////////////////////////
  res->frame=av_frame_alloc();

  if (!res->frame) {
    DMESSAGE("allocating frame");
    goto eframe;
  }

  res->frame->channel_layout=ochannel_layout;
  res->frame->channels=av_get_channel_layout_nb_channels(ochannel_layout);
  res->frame->format=ocodecpar->format;
  res->frame->sample_rate=ocodecpar->sample_rate;

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
  av_frame_free(&res->frame);
eframe:
  swr_close(res->ctx);
  swr_free(&res->ctx);
econtext:
  return -1;
}

void resampler_destroy(resampler_t *res)
{
  av_frame_free(&res->frame);
  swr_close(res->ctx);
  swr_free(&res->ctx);
}

int resampler_apply(resampler_t *res, AVFrame *frame)
{
  int err;

  /////////////////////////////////////////////////////////////////////////////
  err=swr_convert_frame(res->ctx,res->frame,frame);

  if (err<0) {
    DVMESSAGE("converting frame: %s (%d)",av_err2str(err),err);
    goto econvert;
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
econvert:
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
int analyzer_create(analyzer_t *a, inout_t *in, int upsample,
    analyzer_callback_t *cb, void *data)
{
  const param_decode_t *decode=cb->decode?cb->decode(data):NULL;
  int64_t channel_layout=decode?decode->request.channel_layout:-1ll;
  int (*create)(void *, const AVCodecParameters *)=cb?cb->stats.create:NULL;
  AVCodecParameters *icodecpar,ocodecpar;

  /////////////////////////////////////////////////////////////////////////////
  if (in->ai<0) {
    DMESSAGE("input not initialized");
    goto eargs;
  }

  icodecpar=in->fmt.ctx->streams[in->ai]->codecpar;

  /////////////////////////////////////////////////////////////////////////////
  if (!in->audio.ctx) {
    DMESSAGE("audio decoder not opened");
    goto eargs;
  }

  /////////////////////////////////////////////////////////////////////////////
  a->state=ANALYZER_DECODER_SEND_PACKET;
  a->in=in;
  a->data=data;
  a->cb=cb;

  /////////////////////////////////////////////////////////////////////////////
  if (0ll<=channel_layout&&channel_layout!=(int)icodecpar->channel_layout) {
    ocodecpar=*icodecpar;
    ocodecpar.channel_layout=channel_layout;

    ///////////////////////////////////////////////////////////////////////////
    if (resampler_create(&a->normalizer,&ocodecpar,icodecpar)<0) {
      DMESSAGE("creating normalizer");
      goto enormalizer;
    }

    ///////////////////////////////////////////////////////////////////////////
    if (create&&create(data,&ocodecpar)<0) {
      DMESSAGE("creating statistics");
      goto estats;
    }
  }
  else {
    ///////////////////////////////////////////////////////////////////////////
    a->normalizer.ctx=NULL;

    ///////////////////////////////////////////////////////////////////////////
    if (create&&create(data,icodecpar)<0) {
      DMESSAGE("creating statistics");
      goto estats;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  a->pkt=av_packet_alloc();

  if (!a->pkt) {
    DMESSAGE("allocating packet");
    goto epacket;
  }

  /////////////////////////////////////////////////////////////////////////////
  a->frame=av_frame_alloc();

  if (!a->frame) {
    DMESSAGE("allocating frame");
    goto eframe;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (upsample<1)
    a->upsampler.ctx=NULL;
  else {
    ocodecpar=*icodecpar;
    ocodecpar.sample_rate*=upsample;

    if (resampler_create(&a->upsampler,&ocodecpar,icodecpar)<0) {
      DMESSAGE("creating upsampler");
      goto eupsampler;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
  if (a->upsampler.ctx)
    resampler_destroy(&a->upsampler);
eupsampler:
  av_frame_free(&a->frame);
eframe:
  av_packet_free(&a->pkt);
epacket:
  if (cb&&cb->stats.destroy)
    cb->stats.destroy(data);
estats:
  if (a->normalizer.ctx)
   resampler_destroy(&a->normalizer);
enormalizer:
eargs:
  return -1;
}

void analyzer_destroy(analyzer_t *a)
{
  if (a->upsampler.ctx)
    resampler_destroy(&a->upsampler);

  av_frame_free(&a->frame);
  av_packet_free(&a->pkt);

  if (a->cb&&a->cb->stats.destroy)
    a->cb->stats.destroy(a->data);

  if (a->normalizer.ctx)
   resampler_destroy(&a->normalizer);
}

int analyzer_process_frame(analyzer_t *a, AVFrame *frame)
{
  int (*add)(void *, int, AVFrame *)=a->cb?a->cb->stats.add:NULL;
  int err;

  if (a->normalizer.ctx) {
    err=resampler_apply(&a->normalizer,frame);

    if (err<0) {
      DVMESSAGE("applying normalizer: %s (%d)",av_err2str(err),err);
      goto exit;
    }

    if (add) {
      err=add(a->data,0,a->normalizer.frame);

      if (err<0) {
        DVMESSAGE("adding normalized statistics: %s (%d)",av_err2str(err),err);
        goto exit;
      }
    }
  }
  else if (add) {
    err=add(a->data,0,frame);

    if (err<0) {
      DVMESSAGE("adding statistics: %s (%d)",av_err2str(err),err);
      goto exit;
    }
  }

  if (a->upsampler.ctx) {
    err=resampler_apply(&a->upsampler,frame);

    if (err<0) {
      DVMESSAGE("applying upsampler: %s (%d)",av_err2str(err),err);
      goto exit;
    }

    if (add) {
      err=add(a->data,1,a->upsampler.frame);

      if (err<0) {
        DVMESSAGE("adding upsampled statistics: %s (%d)",av_err2str(err),err);
        goto exit;
      }
    }
  }

  err=0;
exit:
#if 0 && defined (FRAME_UNREF) // [
  av_frame_unref(frame);
#endif // ]
  return err;
}

int analyzer_send_packet(analyzer_t *a, AVPacket *pkt)
{
#if 0 // [
  int (*add)(void *, int, AVFrame *)=a->cb?a->cb->stats.add:NULL;
#endif // ]
  int err=-1;

  for (;;) {
    switch (a->state) {
    case ANALYZER_DECODER_SEND_PACKET:
      err=avcodec_send_packet(a->in->audio.ctx,pkt);

#if defined (PACKET_UNREF) // [
      if (pkt)
        av_packet_unref(pkt);
#endif // ]

      switch (err) {
      case AVERROR_INVALIDDATA:
        // Invalid data found when processing input.
        // No respective documentation found. May occur at the end of MP3.
        // Intentional fall-through.
      case 0:
        // 0 on success
        a->state=ANALYZER_DECODER_RECEIVE_FRAME;
        continue;
      default:
        DVMESSAGE("sending packet: %s (%d)",av_err2str(err),err);
        return err;
      }
    case ANALYZER_DECODER_RECEIVE_FRAME:
      err=avcodec_receive_frame(a->in->audio.ctx,a->frame);

      switch (err) {
      case 0:
        // 0: success, a frame was returned
        err=analyzer_process_frame(a,a->frame);
#if defined (FRAME_UNREF) // [
        av_frame_unref(a->frame);
#endif // ]

        if (err<0) {
          DVMESSAGE("processing frame: %s (%d)",av_err2str(err),err);
          goto eloop;
        }

        continue;
      case AVERROR_EOF:
        // we need to flush the encoder.
        a->state=ANALYZER_DECODER_SEND_PACKET;
        return err;
      case AVERROR(EAGAIN):
        // AVERROR(EAGAIN): output is not available in this state - user must
        // try to send new input
        // read the next packet.
        a->state=ANALYZER_DECODER_SEND_PACKET;
        return err;
      default:
        DVMESSAGE("receiving frame: %s (%d)",av_err2str(err),err);
        return err;
      }
    default:
      DMESSAGE("unexpected state");
      goto eloop;
    }
  }

  return 0;
eloop:
  return err;
}

int analyzer_loop(analyzer_t *a)
{
  AVFormatContext *ctx=a->in->fmt.ctx;
  AVPacket *pkt=a->pkt;
  int err;
read:
  err=av_read_frame(ctx,pkt);

  if (err<0)
    goto eof;
  else if ((int)ctx->nb_streams<=pkt->stream_index) {
#if defined (PACKET_UNREF) // [
    av_packet_unref(a->pkt);
#endif // ]
    goto read;
  }

  input_progress(a->in,pkt,stdout);

  if (a->in->ai!=a->pkt->stream_index) {
    // a packet from a stream we're not interested in has to be skipped.
#if defined (PACKET_UNREF) // [
    av_packet_unref(a->pkt);
#endif // ]
    goto read;
  }

  // an audio packet has been read which has to be decoded.
  // the packet is unrefed by analyzer_send_packet().
  err=analyzer_send_packet(a,a->pkt);

  if (ANALYZER_DECODER_SEND_PACKET==a->state) {
    switch (err) {
    case AVERROR(EAGAIN):
      goto read;
    case AVERROR_EOF:
      goto eof;
    default:
      DVMESSAGE("decoding frame: %s (%d)",av_err2str(err),err);
      goto eloop;
    }
  }
eof:
  /////////////////////////////////////////////////////////////////////////////
  // we need to flush the decoder.
  err=analyzer_send_packet(a,NULL);

  if (err<0&&ANALYZER_DECODER_SEND_PACKET<a->state) {
    DVMESSAGE("decoding frame: %s (%d)",av_err2str(err),err);
    goto eloop;
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
eloop:
  return err;
}

///////////////////////////////////////////////////////////////////////////////
int filter_create(filter_t *filter, const AVCodecParameters *ocodecpar,
    const AVCodecParameters *icodecpar, AVRational time_base,
    const char *descr)
{
  // cf. "<ffmpeg>/doc/examples/filtering_audio.c".
  int err=0;
  const enum AVSampleFormat sink_sample_fmts[]
      ={ ocodecpar->format,-1 };
  const int64_t sink_channel_layouts[]
      ={ ocodecpar->channel_layout,-1 };
  const int sink_sample_rates[]
      ={ ocodecpar->sample_rate,-1 };
  char args[512];

  struct {
    const AVFilter *filter;
    AVFilterInOut *in;
  } src;

  struct {
    const AVFilter *filter;
    AVFilterInOut *out;
  } sink;

  /////////////////////////////////////////////////////////////////////////////
  filter->graph=avfilter_graph_alloc();

  if (!filter->graph) {
    DMESSAGE("allocating filter graph");
    err=AVERROR(ENOMEM);
    goto egraph;
  }

  /////////////////////////////////////////////////////////////////////////////
  // buffer audio source: the decoded frames from the decoder will
  // be inserted here.
  src.filter=avfilter_get_by_name("abuffer");

  if (!src.filter) {
    DMESSAGE("audio filter \"abuffer\" not available");
    err=-1;
    goto esrc;
  }

  /////////////////////////////////////////////////////////////////////////////
  snprintf(args, sizeof(args),
      "time_base=%d/%d"
      ":sample_rate=%d"
      ":sample_fmt=%s"
      ":channel_layout=0x%"PRIx64,
      time_base.num,
      time_base.den,
      icodecpar->sample_rate,
      av_get_sample_fmt_name(icodecpar->format),
      icodecpar->channel_layout);
  err=avfilter_graph_create_filter(&filter->ctx.src,src.filter,"in",
      args,NULL,filter->graph);

  if (err<0) {
    DVMESSAGE("creating filter source: %s (%d)",av_err2str(err),err);
    goto esrcctx;
  }

  /////////////////////////////////////////////////////////////////////////////
  // buffer audio sink: to terminate the filter chain.
  sink.filter=avfilter_get_by_name("abuffersink");

  if (!sink.filter) {
    DMESSAGE("audio filter \"abuffersink\" not available");
    err=-1;
    goto esink;
  }

  /////////////////////////////////////////////////////////////////////////////
  err=avfilter_graph_create_filter(&filter->ctx.sink, sink.filter,"out",
      NULL,NULL,filter->graph);

  if (err<0) {
    DVMESSAGE("creating filter sink: %s (%d)",av_err2str(err),err);
    goto esinkctx;
  }

  /////////////////////////////////////////////////////////////////////////////
  err=av_opt_set_int_list(filter->ctx.sink,"sample_fmts",
      sink_sample_fmts,-1,AV_OPT_SEARCH_CHILDREN);

  if (err<0) {
    DVMESSAGE("setting output sample format: %s (%d)",av_err2str(err),err);
    goto esinkargs;
  }

  /////////////////////////////////////////////////////////////////////////////
  err=av_opt_set_int_list(filter->ctx.sink,"channel_layouts",
      sink_channel_layouts,-1,AV_OPT_SEARCH_CHILDREN);

  if (err<0) {
    DVMESSAGE("setting output channel layout: %s (%d)",av_err2str(err),err);
    goto esinkargs;
  }

  /////////////////////////////////////////////////////////////////////////////
  err=av_opt_set_int_list(filter->ctx.sink,"sample_rates",
      sink_sample_rates,-1,AV_OPT_SEARCH_CHILDREN);

  if (err<0) {
    DVMESSAGE("setting output sample rate: %s (%d)",av_err2str(err),err);
    goto esinkargs;
  }

  /////////////////////////////////////////////////////////////////////////////
  /*
   * Set the endpoints for the filter graph. The filter_graph will
   * be linked to the graph described by filters_descr.
   */

  /*
   * The buffer source output must be connected to the input pad of
   * the first filter described by filters_descr; since the first
   * filter input label is not specified, it is set to "in" by
   * default.
   */
  sink.out=avfilter_inout_alloc();

  if (!sink.out) {
    DMESSAGE("allocating outputs");
    goto eoutputs;
  }

  sink.out->name=av_strdup("in");

  if (!sink.out->name) {
    DMESSAGE("duplicating sink name");
    goto esinkname;
  }

  sink.out->filter_ctx=filter->ctx.src;
  sink.out->pad_idx=0;
  sink.out->next= NULL;

  /*
   * The buffer sink input must be connected to the output pad of
   * the last filter described by filters_descr; since the last
   * filter output label is not specified, it is set to "out" by
   * default.
   */
  src.in=avfilter_inout_alloc();

  if (!src.in) {
    DMESSAGE("allocating inputs");
    goto einputs;
  }

  src.in->name=av_strdup("out");

  if (!src.in->name) {
    DMESSAGE("duplicating source name");
    goto esrcname;
  }

  src.in->filter_ctx=filter->ctx.sink;
  src.in->pad_idx=0;
  src.in->next=NULL;

  /////////////////////////////////////////////////////////////////////////////
  err=avfilter_graph_parse_ptr(filter->graph,descr,&src.in,
      &sink.out,NULL);

  if (err<0) {
    DVMESSAGE("parsing: %s (%d)",av_err2str(err),err);
    goto eparse;
  }

  /////////////////////////////////////////////////////////////////////////////
  err=avfilter_graph_config(filter->graph,NULL);

  if (err<0) {
    DVMESSAGE("onfiguring: %s (%d)",av_err2str(err),err);
    goto econfig;
  }

  /////////////////////////////////////////////////////////////////////////////
  avfilter_inout_free(&src.in);
  avfilter_inout_free(&sink.out);

  return 0;
//cleanup:
econfig:
eparse:
esrcname:
  avfilter_inout_free(&src.in);
einputs:
esinkname:
  avfilter_inout_free(&sink.out);
eoutputs:
esinkargs:
esinkctx:
esink:
esrcctx:
esrc:
  avfilter_graph_free(&filter->graph);
egraph:
  return err;
}

void filter_destroy(filter_t *filter)
{
  avfilter_graph_free(&filter->graph);
}

int filter_send_frame(filter_t *filter, AVFrame *frame)
{
  return av_buffersrc_add_frame(filter->ctx.src,frame);
}

int filter_receive_frame(filter_t *filter, AVFrame *frame)
{
  return av_buffersink_get_frame(filter->ctx.sink,frame);
}

///////////////////////////////////////////////////////////////////////////////
int muxer_create(muxer_t *muxer, inout_t *in, inout_t *out, const char *filter)
{
  AVStream *istream=in->fmt.ctx->streams[in->ai];
  AVStream *ostream=out->fmt.ctx->streams[out->ai];

  /////////////////////////////////////////////////////////////////////////////
  muxer->state=MUXER_DECODER_SEND_PACKET;
  muxer->in=in;
  muxer->out=out;

  /////////////////////////////////////////////////////////////////////////////
  muxer->pkt=av_packet_alloc();

  if (!muxer->pkt) {
    DMESSAGE("allocating packet");
    goto epacket;
  }

  if (in->audio.ctx) {
    ///////////////////////////////////////////////////////////////////////////
    muxer->frame=av_frame_alloc();

    if (!muxer->frame) {
    DMESSAGE("allocating frame");
      goto eframe;
    }

    ///////////////////////////////////////////////////////////////////////////
    if (!filter)
      filter="volume";

    if (filter_create(&muxer->filter,ostream->codecpar,istream->codecpar,
        istream->time_base,filter)<0) {
      DMESSAGE("creating filter");
      goto efilter;
    }
  }
  else {
    ///////////////////////////////////////////////////////////////////////////
    muxer->frame=NULL;
    muxer->filter.graph=NULL;
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
  if (muxer->filter.graph)
    filter_destroy(&muxer->filter);
efilter:
  if (muxer->frame)
    av_frame_free(&muxer->frame);
eframe:
  av_packet_free(&muxer->pkt);
epacket:
  return -1;
}

void muxer_destroy(muxer_t *muxer)
{
  if (muxer->filter.graph)
    filter_destroy(&muxer->filter);

  if (muxer->frame)
    av_frame_free(&muxer->frame);

  av_packet_free(&muxer->pkt);
}

int muxer_mux_audio_frame(muxer_t *muxer, AVPacket *pkt)
{
  // cf. "<ffmpeg>/doc/examples/muxing.c".
  audio_t *audio=&muxer->out->audio;
  AVStream *ostream=muxer->out->fmt.ctx->streams[muxer->out->ai];

  // rescale output packet timestamp values from codec to stream timebase.
  av_packet_rescale_ts(pkt,audio->ctx->time_base,ostream->time_base);
  pkt->stream_index=muxer->out->ai;

  // Write the compressed frame to the media file.
  return av_interleaved_write_frame(muxer->out->fmt.ctx,pkt);
}

int muxer_remux_frame(muxer_t *muxer, AVPacket *pkt, int stream_index)
{
  // cf. "<ffmpeg>/doc/examples/remuxing.c".
  AVStream *istream,*ostream;

  istream=muxer->in->fmt.ctx->streams[pkt->stream_index];
  pkt->stream_index=stream_index;
  ostream=muxer->out->fmt.ctx->streams[pkt->stream_index];

  pkt->pts=av_rescale_q_rnd(pkt->pts,istream->time_base,
      ostream->time_base,AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
  pkt->dts=av_rescale_q_rnd(pkt->dts,istream->time_base,
      ostream->time_base,AV_ROUND_NEAR_INF|AV_ROUND_PASS_MINMAX);
  pkt->duration=av_rescale_q(pkt->duration,istream->time_base,
      ostream->time_base);
  pkt->pos=-1;

  return av_interleaved_write_frame(muxer->out->fmt.ctx,pkt);
}

int muxer_send_packet(muxer_t *muxer, AVPacket *pkt)
{
  AVFrame *frame=NULL;
  int err;

  if (!muxer->filter.graph) {
    DMESSAGE("filter not initialized");
    return -1;
  }

  for (;;) {
    switch (muxer->state) {
    case MUXER_DECODER_SEND_PACKET:
      err=avcodec_send_packet(muxer->in->audio.ctx,pkt);

#if defined (PACKET_UNREF) // [
      if (pkt)
        av_packet_unref(pkt);
#endif // ]

      switch (err) {
      case AVERROR_INVALIDDATA:
        // Invalid data found when processing input.
        // No respective documentation found. May occur at the end of MP3.
        // Intentional fall-through.
      case 0:
        // 0 on success
        muxer->state=MUXER_DECODER_RECEIVE_FRAME;
        continue;
      default:
        DVMESSAGE("sending packet: %s (%d)",av_err2str(err),err);
        return err;
      }
    case MUXER_DECODER_RECEIVE_FRAME:
      err=avcodec_receive_frame(muxer->in->audio.ctx,muxer->frame);

      switch (err) {
      case 0:
        // 0: success, a frame was returned
        frame=muxer->frame;
        muxer->state=MUXER_FILTER_SEND_FRAME;
        continue;
      case AVERROR_EOF:
        // we need to flush the encoder.
        frame=NULL;
        muxer->state=MUXER_ENCODER_SEND_FRAME;
        continue;
      case AVERROR(EAGAIN):
        // AVERROR(EAGAIN): output is not available in this state - user must
        // try to send new input
        // read the next packet.
        --muxer->state; // MUXER_DECODER_SEND_PACKET
        return err;
      default:
        DVMESSAGE("receiving frame: %s (%d)",av_err2str(err),err);
        return err;
      }
    case MUXER_FILTER_SEND_FRAME:
      err=filter_send_frame(&muxer->filter,frame);

#if defined (FRAME_UNREF) // [
      if (frame==muxer->frame)
        av_frame_unref(frame);
#endif // ]

      if (err<0) {
        DVMESSAGE("sending frame to filter: %s (%d)",av_err2str(err),err);
        return err;
      }

      muxer->state=MUXER_FILTER_RECEIVE_FRAME;
      continue;
    case MUXER_FILTER_RECEIVE_FRAME:
      err=filter_receive_frame(&muxer->filter,muxer->frame);

      switch (err) {
        case 0:
          frame=muxer->frame;
          muxer->state=MUXER_ENCODER_SEND_FRAME;
          continue;
        case AVERROR(EAGAIN):
          muxer->state=MUXER_DECODER_RECEIVE_FRAME;
          continue;
        case AVERROR_EOF:
          // we should flush the encoder.
          frame=NULL;
          muxer->state=MUXER_ENCODER_SEND_FRAME;
          continue;
        default:
          DVMESSAGE("filter receiving frame: %s (%d)",av_err2str(err),err);
          return err;
      }
    case MUXER_ENCODER_SEND_FRAME:
      err=avcodec_send_frame(muxer->out->audio.ctx,frame);

#if defined (FRAME_UNREF) // [
      if (frame==muxer->frame)
        av_frame_unref(frame);
#endif // ]

      switch (err) {
      case 0:
        muxer->state=MUXER_ENCODER_RECEIVE_PACKET;
        continue;
      default:
        DVMESSAGE("sending frame: %s (%d)",av_err2str(err),err);
        return err;
      }
    case MUXER_ENCODER_RECEIVE_PACKET:
      err=avcodec_receive_packet(muxer->out->audio.ctx,muxer->pkt);

      switch (err) {
      case 0:
        err=muxer_mux_audio_frame(muxer,muxer->pkt);
        fflush(stderr);
#if defined (PACKET_UNREF) // [
        av_packet_unref(muxer->pkt);
#endif // ]

        if (err<0) {
          DMESSAGE("writing packet");
          return err;
        }

        continue;
      case AVERROR_EOF:
#if 0 // [
        // not an error just THE END.
        // everything (decoder and encoder) is flushed.
        DVMESSAGE("sending frame: %s (%d)",av_err2str(err),err);
#else // ] [
        muxer->state=MUXER_DECODER_SEND_PACKET;
#endif // ]
        return err;
      case AVERROR(EAGAIN):
        muxer->state=MUXER_DECODER_RECEIVE_FRAME;
        continue;
      default:
        DVMESSAGE("receiving packet: %s (%d)",av_err2str(err),err);
        return err;
      }
    default:
      DMESSAGE("unexpected state");
      return -1;
    }
  }
}

int muxer_loop(muxer_t *muxer)
{
  int err;
read:
  err=av_read_frame(muxer->in->fmt.ctx,muxer->pkt);

  if (err<0)
    goto eof;
  else if ((int)muxer->in->fmt.ctx->nb_streams<=muxer->pkt->stream_index) {
#if defined (PACKET_UNREF) // [
    av_packet_unref(muxer->pkt);
#endif // ]
    goto read;
  }

  input_progress(muxer->in,muxer->pkt,stdout);

  if (muxer->in->vi==muxer->pkt->stream_index) {
    // a video packet has just to be remuxed.
    err=muxer_remux_frame(muxer,muxer->pkt,muxer->out->vi);
#if defined (PACKET_UNREF) // [
    av_packet_unref(muxer->pkt);
#endif // ]

    switch (err) {
    case 0:
      break;
    case AVERROR(EINVAL):
      DVWARNING("remuxing video frame: %s (%d)",av_err2str(err),err);
      break;
    default:
      if (err<0) {
        DVMESSAGE("remuxing video frame: %s (%d)",av_err2str(err),err);
        goto eloop;
      }

      break;
    }

    goto read;
  }
  else if (muxer->in->ai==muxer->pkt->stream_index&&!muxer->in->audio.ctx) {
    // an audio packet has just to be remuxed.
    err=muxer_remux_frame(muxer,muxer->pkt,muxer->out->ai);
#if defined (PACKET_UNREF) // [
    av_packet_unref(muxer->pkt);
#endif // ]

    switch (err) {
    case 0:
      break;
    case AVERROR(EINVAL):
      DVWARNING("remuxing video frame: %s (%d)",av_err2str(err),err);
      break;
    default:
      if (err<0) {
        DVMESSAGE("remuxing video frame: %s (%d)",av_err2str(err),err);
        goto eloop;
      }

      break;
    }

    goto read;
  }
  else if (muxer->in->ai!=muxer->pkt->stream_index) {
    // a packet from a stream we're not interested in has to be skipped.
#if defined (PACKET_UNREF) // [
    av_packet_unref(muxer->pkt);
#endif // ]
    goto read;
  }

  // an audio packet has been read which has to be decoded, re-encoded,
  // and muxed. the packet is unrefed by muxer_send_packet().
  err=muxer_send_packet(muxer,muxer->pkt);

  if (MUXER_DECODER_SEND_PACKET==muxer->state) {
    switch (err) {
    case AVERROR(EAGAIN):
      goto read;
    case AVERROR_EOF:
      goto eof;
    default:
      DVMESSAGE("re-encoding frame: %s (%d:%d)",
          av_err2str(err),err,muxer->state);
      goto eloop;
    }
  }
eof:
  /////////////////////////////////////////////////////////////////////////////
  if (muxer->in->audio.ctx) {
    // we need to flush the decoder/filter/encoder.
    err=muxer_send_packet(muxer,NULL);

    if (err<0&&MUXER_DECODER_SEND_PACKET<muxer->state) {
      DVMESSAGE("re-encoding frame: %s (%d:%d)",
          av_err2str(err),err,muxer->state);
      goto eloop;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
eloop:
  return err;
}

///////////////////////////////////////////////////////////////////////////////
static iter_vmt_t iter_s16i_vmt;
static iter_vmt_t iter_s32i_vmt;
static iter_vmt_t iter_flti_vmt;
static iter_vmt_t iter_dbli_vmt;
static iter_vmt_t iter_s16p_vmt;
static iter_vmt_t iter_s32p_vmt;
static iter_vmt_t iter_fltp_vmt;
static iter_vmt_t iter_dblp_vmt;
static iter_vmt_t iter_err_vmt;

///////////////////////////////////////////////////////////////////////////////
int iter_first(iter_t *i, const AVFrame *frame)
{
  i->frame=frame;

  switch (frame->format) {
  /////////////////////////////////////////////////////////////////////////////
  case AV_SAMPLE_FMT_NONE:
    DMESSAGE("format AV_SAMPLE_FMT_NONE not supported");
    goto exit;
  // interleaved //////////////////////////////////////////////////////////////
  case AV_SAMPLE_FMT_U8:
    DMESSAGE("format AV_SAMPLE_FMT_U8 not supported");
    goto exit;
  case AV_SAMPLE_FMT_S16:
    i->vmt=&iter_s16i_vmt;
    break;
  case AV_SAMPLE_FMT_S32:
    i->vmt=&iter_s32i_vmt;
    break;
  case AV_SAMPLE_FMT_FLT:
    i->vmt=&iter_flti_vmt;
    break;
  case AV_SAMPLE_FMT_DBL:
    i->vmt=&iter_dbli_vmt;
    break;
  // planar ///////////////////////////////////////////////////////////////////
  case AV_SAMPLE_FMT_U8P:
    DMESSAGE("format AV_SAMPLE_FMT_U8P not supported");
    goto exit;
  case AV_SAMPLE_FMT_S16P:
    i->vmt=&iter_s16p_vmt;
    break;
  case AV_SAMPLE_FMT_S32P:
    i->vmt=&iter_s32p_vmt;
    break;
  case AV_SAMPLE_FMT_FLTP:
    i->vmt=&iter_fltp_vmt;
    break;
  case AV_SAMPLE_FMT_DBLP:
    i->vmt=&iter_dblp_vmt;
    break;
  /////////////////////////////////////////////////////////////////////////////
  case AV_SAMPLE_FMT_NB:
    // Number of sample formats. DO NOT USE if linking dynamically.
    DMESSAGE("format AV_SAMPLE_FMT_NB not supported");
    goto exit;
  default:
    DVMESSAGE("format %d not supported",frame->format);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  i->vmt->first(i);

  /////////////////////////////////////////////////////////////////////////////
  return 0;
exit:
  i->vmt=&iter_err_vmt;
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
static double INT16_SCALE=1.0/PBU_MAXOF(int16_t);
static double INT32_SCALE=1.0/PBU_MAXOF(int32_t);

static void iter_norm(iter_t *i UNUSED, double *x, double *max,
    int ch, double y)
{
  if (x) {
//DVWRITELN("ch: %d y: %.2f",ch,y);
#if ! defined (LIB1770_LFE) // [
    if (ch<3)
#endif // ]
      x[ch]=y;
#if ! defined (LIB1770_LFE) // [
    else if (3<ch)
      x[ch-1]=y;
#endif // ]
  }

  if (max) {
    if (y<0) {
      if (y<-*max)
        *max=-y;
    }
    else if (*max<y)
      *max=y;
  }
}

///////////////////////////////////////////////////////////////////////////////
static void iter_s16i_first(iter_t *i)
{
#if defined (__GNUC__) // [
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-align"
#endif // ]
  i->i.s16.rp=(int16_t *)i->frame->data[0];
#if defined (__GNUC__) // [
#pragma GCC diagnostic pop
#endif // ]
  i->i.s16.mp=i->i.s16.rp+i->frame->channels*i->frame->nb_samples;
}

static int iter_s16i_valid(iter_t *i)
{
  return i->i.s16.rp<i->i.s16.mp;
}

static void iter_s16i_next(iter_t *i)
{
  i->i.s16.rp+=i->frame->channels;
}

static void iter_s16i_norm(iter_t *i, double *x, double *max)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    iter_norm(i,x,max,ch,INT16_SCALE*i->i.s16.rp[ch]);
}

static iter_vmt_t iter_s16i_vmt={
#if defined (PBU_DEBUG) // [
  .id="s16i",
#endif // ]
  .first=iter_s16i_first,
  .valid=iter_s16i_valid,
  .next=iter_s16i_next,
  .norm=iter_s16i_norm,
};

///////////////////////////////////////////////////////////////////////////////
static void iter_s32i_first(iter_t *i)
{
#if defined (__GNUC__) // [
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-align"
#endif // ]
  i->i.s32.rp=(int32_t *)i->frame->data[0];
#if defined (__GNUC__) // [
#pragma GCC diagnostic pop
#endif // ]
  i->i.s32.mp=i->i.s32.rp+i->frame->channels*i->frame->nb_samples;
}

static int iter_s32i_valid(iter_t *i)
{
  return i->i.s32.rp<i->i.s32.mp;
}

static void iter_s32i_next(iter_t *i)
{
  i->i.s32.rp+=i->frame->channels;
}

static void iter_s32i_norm(iter_t *i, double *x, double *max)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    iter_norm(i,x,max,ch,INT32_SCALE*i->i.s32.rp[ch]);
}

static iter_vmt_t iter_s32i_vmt={
#if defined (PBU_DEBUG) // [
  .id="s32i",
#endif // ]
  .first=iter_s32i_first,
  .valid=iter_s32i_valid,
  .next=iter_s32i_next,
  .norm=iter_s32i_norm,
};

///////////////////////////////////////////////////////////////////////////////
static void iter_flti_first(iter_t *i)
{
#if defined (__GNUC__) // [
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-align"
#endif // ]
  i->i.flt.rp=(float *)i->frame->data[0];
#if defined (__GNUC__) // [
#pragma GCC diagnostic pop
#endif // ]
  i->i.flt.mp=i->i.flt.rp+i->frame->channels*i->frame->nb_samples;
}

static int iter_flti_valid(iter_t *i)
{
  return i->i.flt.rp<i->i.flt.mp;
}

static void iter_flti_next(iter_t *i)
{
  i->i.flt.rp+=i->frame->channels;
}

static void iter_flti_norm(iter_t *i, double *x, double *max)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    iter_norm(i,x,max,ch,i->i.flt.rp[ch]);
}

static iter_vmt_t iter_flti_vmt={
#if defined (PBU_DEBUG) // [
  .id="flti",
#endif // ]
  .first=iter_flti_first,
  .valid=iter_flti_valid,
  .next=iter_flti_next,
  .norm=iter_flti_norm,
};

///////////////////////////////////////////////////////////////////////////////
static void iter_dbli_first(iter_t *i)
{
#if defined (__GNUC__) // [
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-align"
#endif // ]
  i->i.dbl.rp=(double *)i->frame->data[0];
#if defined (__GNUC__) // [
#pragma GCC diagnostic pop
#endif // ]
  i->i.dbl.mp=i->i.dbl.rp+i->frame->channels*i->frame->nb_samples;
}

static int iter_dbli_valid(iter_t *i)
{
  return i->i.dbl.rp<i->i.dbl.mp;
}

static void iter_dbli_next(iter_t *i)
{
  i->i.dbl.rp+=i->frame->channels;
}

static void iter_dbli_norm(iter_t *i, double *x, double *max)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    iter_norm(i,x,max,ch,i->i.dbl.rp[ch]);
}

static iter_vmt_t iter_dbli_vmt={
#if defined (PBU_DEBUG) // [
  .id="dbli",
#endif // ]
  .first=iter_dbli_first,
  .valid=iter_dbli_valid,
  .next=iter_dbli_next,
  .norm=iter_dbli_norm,
};

///////////////////////////////////////////////////////////////////////////////
static void iter_s16p_first(iter_t *i)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch) {
#if defined (__GNUC__) // [
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-align"
#endif // ]
    i->p.s16.rp[ch]=(int16_t *)i->frame->data[ch];
#if defined (__GNUC__) // [
#pragma GCC diagnostic pop
#endif // ]
  }

  i->p.s16.mp=i->p.s16.rp[0]+i->frame->nb_samples;
}

static int iter_s16p_valid(iter_t *i)
{
  return i->p.s16.rp[0]<i->p.s16.mp;
}

static void iter_s16p_next(iter_t *i)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    ++i->p.s16.rp[ch];
}

static void iter_s16p_norm(iter_t *i, double *x, double *max)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    iter_norm(i,x,max,ch,INT16_SCALE**i->p.s16.rp[ch]);
}

static iter_vmt_t iter_s16p_vmt={
#if defined (PBU_DEBUG) // [
  .id="s16p",
#endif // ]
  .first=iter_s16p_first,
  .valid=iter_s16p_valid,
  .next=iter_s16p_next,
  .norm=iter_s16p_norm,
};

///////////////////////////////////////////////////////////////////////////////
static void iter_s32p_first(iter_t *i)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch) {
#if defined (__GNUC__) // [
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-align"
#endif // ]
    i->p.s32.rp[ch]=(int32_t *)i->frame->data[ch];
#if defined (__GNUC__) // [
#pragma GCC diagnostic pop
#endif // ]
  }

  i->p.s32.mp=i->p.s32.rp[0]+i->frame->nb_samples;
}

static int iter_s32p_valid(iter_t *i)
{
  return i->p.s32.rp[0]<i->p.s32.mp;
}

static void iter_s32p_next(iter_t *i)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    ++i->p.s32.rp[ch];
}

static void iter_s32p_norm(iter_t *i, double *x, double *max)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    iter_norm(i,x,max,ch,INT32_SCALE**i->p.s32.rp[ch]);
}

static iter_vmt_t iter_s32p_vmt={
#if defined (PBU_DEBUG) // [
  .id="s32p",
#endif // ]
  .first=iter_s32p_first,
  .valid=iter_s32p_valid,
  .next=iter_s32p_next,
  .norm=iter_s32p_norm,
};

///////////////////////////////////////////////////////////////////////////////
static void iter_fltp_first(iter_t *i)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch) {
#if defined (__GNUC__) // [
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-align"
#endif // ]
    i->p.flt.rp[ch]=(float *)i->frame->data[ch];
#if defined (__GNUC__) // [
#pragma GCC diagnostic pop
#endif // ]
  }

  i->p.flt.mp=i->p.flt.rp[0]+i->frame->nb_samples;
}

static int iter_fltp_valid(iter_t *i)
{
  return i->p.flt.rp[0]<i->p.flt.mp;
}

static void iter_fltp_next(iter_t *i)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    ++i->p.flt.rp[ch];
}

static void iter_fltp_norm(iter_t *i, double *sample, double *max)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    iter_norm(i,sample,max,ch,*i->p.flt.rp[ch]);
}

static iter_vmt_t iter_fltp_vmt={
#if defined (PBU_DEBUG) // [
  .id="fltp",
#endif // ]
  .first=iter_fltp_first,
  .valid=iter_fltp_valid,
  .next=iter_fltp_next,
  .norm=iter_fltp_norm,
};

///////////////////////////////////////////////////////////////////////////////
static void iter_dblp_first(iter_t *i)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch) {
#if defined (__GNUC__) // [
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-align"
#endif // ]
    i->p.dbl.rp[ch]=(double *)i->frame->data[ch];
  }

  i->p.dbl.mp=i->p.dbl.rp[0]+i->frame->nb_samples;
}

static int iter_dblp_valid(iter_t *i)
{
  return i->p.dbl.rp[0]<i->p.dbl.mp;
}

static void iter_dblp_next(iter_t *i)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    ++i->p.dbl.rp[ch];
}

static void iter_dblp_norm(iter_t *i, double *x, double *max)
{
  int ch;

  for (ch=0;ch<i->frame->channels;++ch)
    iter_norm(i,x,max,ch,*i->p.dbl.rp[ch]);
}

static iter_vmt_t iter_dblp_vmt={
#if defined (PBU_DEBUG) // [
  .id="dblp",
#endif // ]
  .first=iter_dblp_first,
  .valid=iter_dblp_valid,
  .next=iter_dblp_next,
  .norm=iter_dblp_norm,
};

///////////////////////////////////////////////////////////////////////////////
static void iter_err_first(iter_t *i UNUSED)
{
  DMESSAGE("invalid frame iterator");
}

static int iter_err_valid(iter_t *i UNUSED)
{
  DMESSAGE("invalid frame iterator");

  return 0;
}

static void iter_err_next(iter_t *i UNUSED)
{
  DMESSAGE("invalid frame iterator");
}

static void iter_err_norm(iter_t *i UNUSED, double *x UNUSED,
    double *max UNUSED)
{
  DMESSAGE("invalid frame iterator");
}

static iter_vmt_t iter_err_vmt={
#if defined (PBU_DEBUG) // [
  .id="err",
#endif // ]
  .first=iter_err_first,
  .valid=iter_err_valid,
  .next=iter_err_next,
  .norm=iter_err_norm,
};

///////////////////////////////////////////////////////////////////////////////
typedef struct {
  param_decode_t decode;

  struct {
    double samplepeak;
    double truepeak;
    lib1770_stats_t *momentary;
  } stats;

  struct {
    lib1770_pre_t *pre;
  } filter;

  struct {
    lib1770_block_t *momentary;
  } block;
} param_t;

const param_decode_t *param_decode(const void *data)
{
  return &((const param_t *)data)->decode;
}

int param_stats_create(void *data, const AVCodecParameters *codecpar)
{
  param_t *param=data;

  /////////////////////////////////////////////////////////////////////////////
  param->stats.samplepeak=0.0;
  param->stats.truepeak=0.0;

  /////////////////////////////////////////////////////////////////////////////
  param->stats.momentary=lib1770_stats_new();

  if (!param->stats.momentary) {
    DMESSAGE("creating momentary statistics");
    goto estats;
  }

  /////////////////////////////////////////////////////////////////////////////
  param->filter.pre
      =lib1770_pre_new_lfe(codecpar->sample_rate,codecpar->channels,3);

  if (!param->filter.pre) {
    DMESSAGE("creating filter");
    goto epre;
  }

  /////////////////////////////////////////////////////////////////////////////
  param->block.momentary=lib1770_block_new(codecpar->sample_rate,400.0,4);

  if (!param->block.momentary) {
    DMESSAGE("creating momentary block");
    goto eblock;
  }

  lib1770_block_add_stats(param->block.momentary,param->stats.momentary);
  lib1770_pre_add_block(param->filter.pre,param->block.momentary);

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
  lib1770_block_close(param->block.momentary);
eblock:
  lib1770_pre_close(param->filter.pre);
epre:
  lib1770_stats_close(param->stats.momentary);
estats:
  return -1;
}

void param_stats_destroy(void *data)
{
  param_t *param=data;

  lib1770_block_close(param->block.momentary);
  lib1770_pre_close(param->filter.pre);
  lib1770_stats_close(param->stats.momentary);
}

int param_stats_add(void *data, int upsampled, AVFrame *frame)
{
  param_t *param=data;
  lib1770_sample_t sample={ 0 };
  iter_t i;

  if (upsampled) {
    for (iter_first(&i,frame);i.vmt->valid(&i);i.vmt->next(&i))
      i.vmt->norm(&i,NULL,&param->stats.truepeak);
  }
  else {
    for (iter_first(&i,frame);i.vmt->valid(&i);i.vmt->next(&i)) {
      i.vmt->norm(&i,sample,&param->stats.samplepeak);
      lib1770_pre_add_sample(param->filter.pre,sample);
    }
  }

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
  analyzer_callback_t analyzer_callback={
    .decode=param_decode,
    .stats.create=param_stats_create,
    .stats.destroy=param_stats_destroy,
    .stats.add=param_stats_add,
  };

  param_t param={
    .decode.request.sample_fmt=AV_SAMPLE_FMT_NONE,
    .decode.request.channel_layout=-1ll,
    .decode.drc.enabled=1,
    .decode.drc.scale=0.0,
  };

  int err=1;
#if defined (FF_ZERO_INITIALIZER) // [
  inout_t in={ 0 };
  analyzer_t analyzer={ 0 };
  inout_t out={ 0 };
  muxer_t muxer={ 0 };
#else // ] [
  inout_t in;
  analyzer_t analyzer;
  inout_t out;
  muxer_t muxer;
#endif // ]

  int upsample=0;
  int apply=0;
  char *targetdir=NULL;
  char *ext=NULL;
  int c;
  char *source;
  char *sp1,*sp2;
  char target[PATH_MAX],*tp=target;
  enum AVSampleFormat sample_fmt=AV_SAMPLE_FMT_S32;
  double norm=-23.0,loudness,gain,volume;
  char filter[64];

#if ! defined (FF_ZERO_INITIALIZER) // [
  /////////////////////////////////////////////////////////////////////////////
  memset(&in,0,sizeof in);
  memset(&analyzer,0,sizeof analyzer);
  memset(&out,0,sizeof out);
  memset(&muxer,0,sizeof muxer);
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  for (;;) {
    c=getopt_long(argc,argv,"sato:e:b:",NULL,NULL);

    if (c<0)
      break;

    switch (c) {
    case 's':
      param.decode.request.channel_layout=AV_CH_LAYOUT_STEREO;
      continue;
    case 'a':
      apply=1;
      continue;
    case 't':
      upsample=4;
      continue;
    case 'o':
      targetdir=optarg;
      continue;
    case 'e':
      ext=optarg;
      continue;
    case 'b':
      sample_fmt=atoi(optarg)<24?AV_SAMPLE_FMT_S16:AV_SAMPLE_FMT_S32;
      continue;
    default:
      if ('c'==optopt) {
        DVMESSAGE("option -%c requires an argument",optopt);
      }
      else if (isprint(optopt)) {
        DVMESSAGE("unknown option -%c",optopt);
      }
      else {
        DVMESSAGE("unknown option character \\x%x",optopt);
      }

      goto eargs;
    }
  }

  if (optind==argc) {
    DMESSAGE("missing argument");
    goto eargs;
  }

  source=argv[optind++];
  fprintf(stdout,"source: \"%s\"\n",source);
  fflush(stdout);

  if (targetdir) {
    tp+=strlen(targetdir);
    memcpy(target,argv[2],tp-target);
    *tp++='/';
    sp1=basename(source);
    sp2=basenamex(sp1,".");
    memcpy(tp,sp1,sp2-sp1);
    tp+=sp2-sp1;
    strcpy(tp,ext?ext:sp2);
    fprintf(stdout,"target: \"%s\"\n",target);
    fflush(stdout);
  }

  /////////////////////////////////////////////////////////////////////////////
#if (LIBAVFILTER_VERSION_MAJOR<7) // [
  avfilter_register_all();
#endif // ]
#if (LIBAVFORMAT_VERSION_MAJOR<58) // [
  av_register_all();
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  if (input_create(&in,source,&param.decode)<0) {
    DMESSAGE("creating input");
    goto einput;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (analyzer_create(&analyzer,&in,upsample,&analyzer_callback,&param)<0) {
    DMESSAGE("creating analyzer");
    goto eanalyzer;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (analyzer_loop(&analyzer)<0) {
    DMESSAGE("analyzer looping");
    goto ealoop;
  }

  /////////////////////////////////////////////////////////////////////////////
  input_close(&in);
  loudness=lib1770_stats_get_mean(param.stats.momentary,-10.0);
  gain=norm-loudness;
  volume=LIB1770_DB2Q(gain);
  fflush(stderr);
  fprintf(stdout,"loudness: %1.2f gain: %1.2f volume: %1.2f"
			" samplepeak: %1.2f",
      loudness,gain,volume,param.stats.samplepeak);

  if (1<upsample)
    fprintf(stdout," truepeak: %1.2f\n",param.stats.truepeak);
  else
    fputc('\n',stdout);

  fflush(stdout);

  if (input_open(&in,source,apply?&param.decode:NULL)<0) {
    DMESSAGE("re-opening input");
    goto eopen;
  }

  if (targetdir) {
    ///////////////////////////////////////////////////////////////////////////
    if (output_create(&out,target,&analyzer,sample_fmt)<0) {
      DMESSAGE("creating output");
      goto eoutput;
    }

    ///////////////////////////////////////////////////////////////////////////
    if (apply)
      snprintf(filter,sizeof filter,"volume=%1.2f",volume);

    if (muxer_create(&muxer,&in,&out,apply?filter:NULL)<0) {
      DMESSAGE("creating re-encoder");
      goto emuxer;
    }

    ///////////////////////////////////////////////////////////////////////////
    if (muxer_loop(&muxer)<0) {
      DMESSAGE("re-encoder looping");
      goto emloop;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  fflush(stderr);
  fprintf(stdout,"done.\n");
  fflush(stdout);
  err=0;
//cleanup:
emloop:
  if (targetdir)
    muxer_destroy(&muxer);
emuxer:
  if (targetdir)
    output_destroy(&out);
eoutput:
eopen:
ealoop:
  analyzer_destroy(&analyzer);
eanalyzer:
  input_destroy(&in);
einput:
eargs:
  return err;
}
